/* $Id: Backup.c,v 1.58 2008/06/12 00:42:01 arpruss Exp $ */
#include <PalmOS.h>
#include <VFSMgr.h>
#include <FSLib.h>
#include <PceNativeCall.h>
#include <68k/system/PmPalmOSNVFS.h>
#include "Backup.h"
#include "crc32.h"

#define DEFAULT_ENCRYPTION 'AES1' // 'TEAn' // 'AES1' // 'DES_'

#define THRESHOLD_1000   1005  // 1/1000s battery threshold needed

//#define DIM

//#include "wave.c"

#define DBCACHE_FREE_PERCENTAGE  50
#define DBCACHE_CHUNK_PERCENTAGE 30

#define SPARE_SPACE 4096

#define MAX_RETRY_COUNT 3
//#define RANDOM_FAILURE

#define BUFFER_SIZE 15000 // 60000 // 15000 // Does not need to be big as the ZVFS*() routines have a big buffer

#define LOG_BUFFER_SIZE 1024 //16384

#define MY_FONT     0x80

#define IS_SD( m ) ( (m) == expMediaType_SecureDigital || \
                      (m) == expMediaType_MultiMediaCard )

#define IS_INTERNAL_DRIVE( m ) ( (m) == 'TFFS' || (m) == 'HDSK' )

#define SAME_BASE( f, g ) ( ( ( ( Char* )( f ) )[ 0 ] == ( ( Char* )( g ) )[ 0 ] ) && SameBaseFileName( (f), (g) ) )

#define CRYPT( v ) ( ( ( v ) == volRefHidden ) ? NULL : &crypt )

typedef struct {
    UInt32  crc;
    UInt32  size;
    Int32   srcIndex;
    Boolean verifyOnly;
} RetryType;

//static Boolean    updatingEncrypted = false;
Boolean       locked = false;
static CryptInfo  crypt = { NULL, NULL };
static Int8 compressLevel;
static Char**     namesOutDir = NULL;
static Int32      countOutDir = 0;
static Char**     namesInDir  = NULL;
static Int32      countInDir  = 0;
static RetryType* retries = NULL;
static Int32      countRetries = 0;
static    VolumeInfoType hiddenInfo;
static    VolumeInfoType cardInfo;
static    VolumeInfoType ignoreInfo;
static    UInt16        volRefSrc;
static    UInt16        volRefDest;
static    UInt16        volRefCard;
static    UInt16        volRefHidden;
static Boolean errorInBackup = false;
#ifdef DIM
static    Int16         oldBrightness = -1;
#endif
static    Boolean       disabledNotifications = false;
Boolean nvfsOff = false;

static Boolean noLog = false;

static FileRef          logRef;
static Char*            logBuffer = NULL;
static UInt16           inLogBuffer;

static Char*            buffer1;
static Char*            buffer2;

static const UInt32* notifications;
static UInt16 numNotifications;

static const UInt32 moreNotifications[] = {
    sysNotifyVolumeUnmountedEvent,
    sysNotifyVolumeMountedEvent,
    '-pdb',
    'Batt',
    'BTdn',
    'BTup',
    'CNNR',
    'DmTx',
    'DsOn',
    'ExAs',
    'GAny',
    'GrFn',
    'Hchg',
    'MyKL',
    'PcLK',
    'PinE',
    'RChg',
    'REMq',
    'Rchg',
    'SCam',
    'SChg',
    'SnC5',
    'SnC7',
    'WFSU',
    'WiDn',
    'WiLC',
    'WiUp',
    'WiSc',
    'ZiCL',
    'arun',
    'blth',
    'brtP',
    'bsnt',
    'btdy',
    'btsn',
    'btsv',
    'cbar',
    'char',
    'cncp',
    'crdi',
    'crdo',
    'dbcr',
    'dbdr',
    'dbmn',
    'dbs-',
    'ecna',
    'ecnd',
    'evtg',
    'g2en',
    'g2tu',
    'gMgC',
    'hapl',
    'hapq',
    'hede',
    'hlpr',
    'hots',
    'htaD',
    'hvch',
    'lazy',
    'locL',
    'locM',
    'locN',
    'locc',
    'neti',
    'pdmE',
    'pset',
    'retk',
    'rfsL',
    'rstf',
    'scrd',
    'scrs',
    'sdpa',
    'slp!',
    'slpq',
    'sync',
    'terr',
    'time',
    'tmgr',
    'tpin',
    'unlk',
//    'volm',
//    'volu',
    'worm',
};

static const UInt32 fewerNotifications[] = {
    sysNotifyDBDirtyEvent,				//'dbdr',
	sysNotifyDeleteProtectedEvent,		//'-pdb',
    sysNotifyMenuCmdBarOpenEvent,		//'cbar',
    sysNotifyProcessPenStrokeEvent,		//'hpps'
    sysNotifyVirtualCharHandlingEvent,	//'hvch',
    sysNotifyEventDequeuedEvent,		//'hede',
    sysNotifyIdleTimeEvent,				//'hidl'
    sysNotifySleepRequestEvent,			//'slpq',
    sysNotifySleepNotifyEvent,			//'slp!',
    sysNotifyEarlyWakeupEvent,			//'worm',
    sysNotifyLateWakeupEvent,			//'lazy',
    'scrs',								// sysNotifyDisplayResizedEvent,
    sysNotifyDisplayChangeEvent,		//'scrd',
};

#ifndef memHeapFlagReadOnly
# define memHeapFlagReadOnly 0x0001
#endif

#define memMaxChunkAllocSize  65505ul
#define memNewChunkFlagAllowLarge 0x1000
#define dmOrphanOwnerID            3

static void PauseReset( void );

MemHandle AllocateChunk( UInt32 size, Boolean dynamic, Boolean moveable )
{
    UInt16 i;
    UInt16 n;
    UInt16 card;
    UInt16 moveableAttr;

    if ( moveable )
        moveableAttr = 0;
    else
        moveableAttr = memNewChunkFlagNonMovable;

    for ( card = 0 ; card < 1 ; card++ ) {
        n = MemNumHeaps( card );

        for ( i = 0 ; i < n ; i++ ) {
             UInt16 heapID;
             UInt16 flags;

             heapID = MemHeapID( card, i );

             flags  = MemHeapFlags( heapID );

             if ( ! ( flags & memHeapFlagReadOnly ) && !! MemHeapDynamic( heapID ) == !! dynamic ) {
                 MemHandle h;
                 h = MemChunkNew( heapID, size, dmOrphanOwnerID |
                                            memNewChunkFlagAllowLarge |
                                            moveableAttr );
                 if ( h != NULL )
                     return h;
             }
        }
    }
    return NULL;
}


void* MemPtrNewBig( UInt32 size )
{
//    if ( size < 65000 )
//        return MemPtrNew( size );
    return AllocateChunk( size, true, false );
}

Err DisableNotification( SysNotifyParamType *np )
{
    /* disable notification */
    np->handled    = 0xFF;
    np->notifyType = 'NVbk';
    return errNone;
}


void Register( UInt32 type, SysNotifyProcPtr f, Int8 priority )
{
    UInt16    card;
    LocalID   id;

    if ( errNone == SysCurAppDatabase( &card, &id ) )
        SysNotifyRegister( card, id, type, f, priority, NULL );
}


void Unregister( UInt32 type, Int8 priority )
{
    UInt16    card;
    LocalID   id;

    if ( errNone == SysCurAppDatabase( &card, &id ) )
        SysNotifyUnregister( card, id, type, priority );
    else {
        Message( FATALERROR, "Fatal error in Unregister()...  Resetting..." );
        PauseReset();
    }
}


static void LogRollover( UInt16 vol, const Char* dir )
{
    Char* logPath;
    Char* logPath2;
    FileRef ref;

    logPath = MemPtrNew( StrLen( dir ) + 1 + sizeof( logName ) );
    if ( logPath == NULL )
        return;
    StrPrintF( logPath, "%s/%s", dir, logName );

    if ( errNone == VFSFileOpen( vol, logPath, vfsModeRead, &ref ) ) {
        UInt32 size;
        if ( errNone != VFSFileSize( ref, &size ) )
            size = 0;
        VFSFileClose( ref );
        if ( LOG_ROLLOVER_SIZE <= size ) {
            logPath2 = MemPtrNew( StrLen( dir ) + 1 + sizeof( logName2 ) );
            if ( logPath2 != NULL ) {
                StrPrintF( logPath2, "%s/%s", dir, logName2 );
                VFSFileDelete( vol, logPath2 );
                VFSFileRename( vol, logPath2, logPath );
                MemPtrFree( logPath2 );
            }
        }
    }
    MemPtrFree( logPath );
}



static void InitLog( UInt16 vol, const Char* dir )
{
    Err err;

    logBuffer = MemPtrNew( LOG_BUFFER_SIZE );

    StrPrintF( logBuffer, "%s/%s", dir, oldLogName );

    VFSFileRename( vol, logBuffer, logName );

    StrPrintF( logBuffer, "%s/%s", dir, logName );

    LogRollover( vol, dir );

    err = VFSFileCreate( vol, logBuffer );

    if ( errNone != VFSFileOpen( vol, logBuffer, ( err == errNone ) ? vfsModeWrite : vfsModeReadWrite, &logRef ) ) {
        MemPtrFree( logBuffer );
        logBuffer = NULL;
        if ( err != errNone )
            Message( ERROR, "Error opening log..." );
    }

    inLogBuffer = 0;

    VFSFileSeek( logRef, vfsOriginEnd, 0 );
}


static void AddLog( const Char* s, UInt16 length )
{
    while ( 0 < length ) {
        UInt16 space;

//            VFSFileWrite( logRef, length, s, 0 );
//            return;

        space = LOG_BUFFER_SIZE - inLogBuffer;

        if ( length <= space )
            space = length;

        if ( 0 < space ) {
            MemMove( logBuffer + inLogBuffer, s, space );
        }

        inLogBuffer += space;
        s           += space;
        length      -= space;

        if ( LOG_BUFFER_SIZE <= inLogBuffer ) {
            VFSFileWrite( logRef, inLogBuffer, logBuffer, 0 );
            inLogBuffer = 0;
        }
    }
}


static void CloseLog( void )
{
    if ( logBuffer != NULL ) {
        if ( 0 < inLogBuffer ) {
            VFSFileWrite( logRef, inLogBuffer, logBuffer, 0 );
            inLogBuffer = 0;
        }
        MemPtrFree( logBuffer );
        logBuffer   = NULL;
        VFSFileClose( logRef );
    }
}


static void LogMessage( const Char* s )
{
    if ( logBuffer == NULL )
        return;
    if ( *s == '\b' )
        s++;
    AddLog( s, StrLen( s ) );
    AddLog( "\r\n", 2 );
}

void Pause( UInt16 n )
{
    UInt32 t;

    t = TimGetTicks() + n * SysTicksPerSecond();

    while ( TimGetTicks() < t );
}

static void PauseReset( void )
{
    Message( INFO, "\b  Resetting." );
    CloseLog();
    Pause( 3 );
    SysReset();
}


void LoadMessageFont( void )
{
    MemHandle h;
    UInt32    value;

    if ( scheduled && errNone == FtrGet( appCreator, FTR_UI_SWITCH, &value ) &&
         ! value )
        return;

    h = DmGetResource( 'nfnt', 0 );

    if ( h != NULL ) {
        FntDefineFont( MY_FONT, MemHandleLock( h ) );
        MemHandleUnlock( h );
        DmReleaseResource( h );
    }
}


void Message( MessageType m, const Char* s )
{
    static Coord lastX = 0;
    FontID oldFont;
    RectangleType display; // = { { 0, 12 }, { 160, 160-12 } };
    RectangleType vacated;
    RGBColorType  c;
    Coord         maxX;
    Coord         maxY;
//    static UInt16  wavePosition = 0;

    EvtFlushKeyQueue();
    EvtFlushPenQueue();

    if ( ! noLog ) {
        LogMessage( s );
    }

    if ( locked && m != ERROR && m != FATALERROR )
        s = "[private]";

    WinGetDisplayExtent( &maxX, &maxY );
    WinGetDrawWindowBounds( &display );
    if ( maxX < display.topLeft.x + display.extent.x )
        display.extent.x = maxX - display.topLeft.x;
    if ( maxY < display.topLeft.y + display.extent.y )
        display.extent.y = maxY - display.topLeft.y;
    WinDisplayToWindowPt( &display.topLeft.x, &display.topLeft.y );
    display.topLeft.y += 12;
    display.extent.y  -= 12; // display.topLeft.y;
//display.extent.y = 20;

    WinPushDrawState();
    switch( m ) {
        case ERROR:
            c.r = 200;
            c.g = 0;
            c.b = 0;
            break;

        case FATALERROR:
            c.r = 220;
            c.g = 0;
            c.b = 0;
            break;

        case TO_SERVER:
            c.r = 0;
            c.g = 0;
            c.b = 128;
            break;

        case FROM_SERVER:
        case IMPORTANT:
            c.r = 0;
            c.g = 128;
            c.b = 0;
            break;

#ifdef DEBUG
        case DEBUG:
            c.r = 0;
            c.g = 0;
            c.b = 255;
            break;
#endif

//        case INFO:
        default:
            c.r = 0;
            c.g = 0;
            c.b = 0;
            break;
    }

    WinSetTextColorRGB( &c, NULL );
//    WinSetBackColorRGB( &wave[ wavePosition ], NULL );
//    wavePosition = ( wavePosition + 1 ) % NUM_WAVE_VALUES;

    oldFont = FntSetFont( MY_FONT );

    if ( s[ 0 ] == '\b' ) {
        s++;
    }
    else {
        lastX = 0;
        WinScrollRectangle( &display, winUp, FntLineHeight(), &vacated );
        WinEraseRectangle( &vacated, 0 );
    }


    WinDrawChars( s, StrLen( s ),
        lastX + display.topLeft.x, display.topLeft.y + display.extent.y - FntLineHeight() );

    lastX += FntCharsWidth( s, StrLen( s ) );

    FntSetFont( oldFont );
    WinPopDrawState();

    if ( *s == ':' )
        Pause( 2 );
}


void ProgressBar( Int16 barNumber, UInt32 current, UInt32 total )
{
    RectangleType r = { { 0, 0 }, { 160, 6 } };
    Coord         delta;
    Coord         x,y;
    RGBColorType  barColor;

    WinGetDisplayExtent( &x,&y );
    
    if ( total == 0 )
        total = 1;

    delta = x * current / total;
    r.extent.x = delta;

    if ( barNumber == -1 ) {
        r.topLeft.y = 0;
        r.extent.y  = 12;
    }
    else {
        r.topLeft.y = 6 * barNumber;
    }

    WinPushDrawState();

    barColor.r = 0;
    barColor.g = current * 255 / total;
    barColor.b = 255 - current * 255 / total;
    WinSetForeColorRGB( &barColor, NULL );

    WinDrawRectangle( &r, 0 );

    r.topLeft.x = delta;
    r.extent.x  = x - delta;

    barColor.r = 255;
    barColor.g = 255;
    barColor.b = 255;

    WinSetBackColorRGB( &barColor, NULL );

    WinEraseRectangle( &r, 0 );

    WinPopDrawState();
}


void ClearScreen( void )
{
    RectangleType display;// = { { 0, 0 }, { 160, 160 } };

    WinGetDrawWindowBounds( &display );
    display.topLeft.x = 0;
    display.topLeft.y = 0;

//    WinGetDisplayExtent( &display.extent.x, &display.extent.y );

    WinEraseRectangle( &display, 0 );
}



static Boolean CopyFileRef( ZFile* out, ZFile* in, UInt32 inSize )
{
    UInt16  didRead;
    Err     err;
    Boolean success;
    UInt32  readSoFar;

    success = true;

    readSoFar = 0;

    do {
        ProgressBar( 1, readSoFar * 2, inSize * 2 );

        err = ZVFSRead( in, BUFFER_SIZE, buffer1, &didRead );

        if ( err != errNone && err != vfsErrFileEOF ) {
            Message( ERROR, "Error reading." );
            Pause( 1 );
            success = false;
            break;
        }

        if ( didRead == 0 ) {
            break;
        }

        ProgressBar( 1, readSoFar * 2 + didRead, inSize * 2 );

        if ( errNone != ZVFSWrite( out, didRead, buffer1 ) ) {
            Message( ERROR, "Error writing." );
            Pause( 1 );
            success = false;
            break;
        }
        else {
            if ( volRefSrc == volRefHidden && ! out->compressed ) {
                out->crc32 = CRC32( out->crc32, buffer1, didRead );
            }
        }

        readSoFar += didRead;
    } while ( err == errNone );

    if ( success )
        ProgressBar( 1, 1, 1 );

    return success;
}


// closes ref2
static Boolean SameFile( UInt16 vol1, Char* name1, ZFile* ref2, UInt32 size2 )
{
    UInt32 size1;
    ZFile file1;

    if ( errNone != ZVFSOpenRead( vol1, name1, &file1, CRYPT( vol1 ) ) ) {

        ZVFSFileClose( ref2, false );

        return false;
    }

    if ( ! pref.slowUpdate && ( file1.compressed || ref2->compressed ) ) {
        Err err;
        UInt32 size1 = 0xFFFFFFFF;
        UInt32 crc1;
        UInt32 crc2;
        UInt32 date1;
        UInt32 date2;

        err = ZVFSStats( &file1, &date1, &size1, NULL );

        if ( err == errNone && size1 != size2 ) {
            err = 1;
        }

        if ( err == errNone ) {
            err = ZVFSStats( ref2, &date2, NULL, NULL );
        }

        if ( err == errNone && date1 != date2 ) {
            err = 1;
        }

        if ( err == errNone && ! ( pref.trustDates && volRefSrc == volRefHidden ) ) {
            Err err1;
            Err err2;

            /* this closes the files */
            err1 = ZVFSStats( &file1, NULL, NULL, &crc1 );
            err2 = ZVFSStats( ref2, NULL, NULL, &crc2 );

            return err1 == errNone && err2 == errNone && crc1 == crc2;
        }
        else {
            ZVFSFileClose( &file1, false );
            ZVFSFileClose( ref2, false );

            return err == errNone;
        }
    }

    if ( errNone != ZVFSFileSize( &file1, &size1 ) ) {

        ZVFSFileClose( &file1, false );
        ZVFSFileClose( ref2, false );

        return false;
    }

    if ( size1 != size2 ) {

        ZVFSFileClose( &file1, false );
        ZVFSFileClose( ref2, false );

        return false;
    }

    while ( 0 < size1 ) {
        Err err;
        UInt32 size;

        if ( size1 < BUFFER_SIZE )
            size = size1;
        else
            size = BUFFER_SIZE;

        err = ZVFSRead( &file1, size, buffer1, NULL );

        if ( err != errNone ) {
            break;
        }

        err = ZVFSRead( ref2, size, buffer2, NULL );

        if ( err != errNone ) {
            break;
        }

        if ( 0 != MemCmp( buffer1, buffer2, size ) ) {
            break;
        }

        size1 -= size;
    }

    ZVFSFileClose( &file1, false );
    ZVFSFileClose( ref2, false );
//    ZVFSRewind( ref2 );

    return 0 == size1;
}




static Boolean SameBaseFileName( Char* a, Char* b )
{
    UInt16 length1;
    UInt16 length2;

    length1 = StrLen( a );

    if ( 3 < length1 && ! StrCompare( a + length1 - 3, ".gz" ) )
        length1 -= 3;

    length2 = StrLen( b );

    if ( 3 < length2 && ! StrCompare( b + length2 - 3, ".gz" ) )
        length2 -= 3;

    if ( length1 != length2 )
        return false;

    return 0 == StrNCompare( a, b, length1 );
}



Err FileDeleteBoth( UInt16 destVol, Char* destPath )
{
    static Char name[ MAX_FILENAME + 64 ];
    UInt16 length;
    Err    err1;
    Err    err2;

    length = StrLen( destPath );

    err1 = VFSFileDelete( destVol, destPath );

    if ( 3 < length && ! StrCompare( destPath + length - 3, ".gz" ) ) {
        StrNCopy( name, destPath, length - 3 );
        name[ length - 3 ] = 0;
    }
    else {
        StrPrintF( name, "%s.gz", destPath );
    }
    err2 = VFSFileDelete( destVol, name );

    if ( err1 != errNone && err2 != errNone )
        return err1;
    else
        return errNone;
}



static Err GetFreeSpace( UInt16 vol, UInt32* freeSpaceP )
{
    Err err;
    UInt32 usedSpace;
    UInt32 totalSpace;

    err = VFSVolumeSize( vol, &usedSpace, &totalSpace );

    if ( err == errNone ) {
        *freeSpaceP = totalSpace - usedSpace;
//        *freeSpaceP = 10000;
    }
    return err;
}



static Boolean IsEssential( Char* name )
{
    return false; // TO DO!
}


static void DeleteMatching( UInt16 vol, Char* destDir, Char* inName )
{
    Int32 i;
    static Char name0[ MAX_FILENAME + 64 ];

    for ( i = 0 ; i < countOutDir ; i++ ) {
         if ( SAME_BASE( inName, namesOutDir[ i ] ) ) {
             Err err;

             StrPrintF( name0, "%s/%s", destDir, namesOutDir[ i ] );

             err = VFSFileDelete( vol, name0 );

             if ( err == errNone ) {
                 namesOutDir[ i ][ 0 ] = 0;
             }
         }
    }
}


static void EnsureSpace( UInt16 vol, Char* destDir, UInt32 spaceNeeded, Int32 startIndex )
{
    UInt32 freeSpace;
    Int32  i;
    Boolean didMessage = false;

    spaceNeeded += SPARE_SPACE;
    i = startIndex;

    while ( errNone == GetFreeSpace( vol, &freeSpace ) && freeSpace < spaceNeeded &&
            i < countInDir ) {

        if ( IsEssential( namesInDir[ i ] ) )
            continue;

        if ( ! didMessage ) {
            Message( INFO, "Cleaning up data to make space..." );
            didMessage = true;
        }

        DeleteMatching( vol, destDir, namesInDir[ i ] );
        i++;
    }

    i = startIndex;
    while ( errNone == GetFreeSpace( vol, &freeSpace ) && freeSpace < spaceNeeded &&
            i < countInDir ) {

        if ( ! didMessage ) {
            Message( INFO, "Cleaning up data to make space..." );
            didMessage = true;
        }

        DeleteMatching( vol, destDir, namesInDir[ i ] );
        i++;
    }

    if ( didMessage ) {
        Char m[30];

        StrPrintF( m, "Free space: %ld.", freeSpace );
        Message( INFO, m );
    }
}



static void CopyFile( UInt16 destVol, Char* destDir, Char* destFileName, UInt16 srcVol,
    Char* srcPath, Int32 srcIndex,
    Boolean doAll, Boolean useTemp )
{
    Err err;
    ZFile src;
    ZFile dest;
    UInt32  size;
    UInt32  mdate;
    Int32   i;
    Boolean success;
    static Char    destName[ MAX_FILENAME + 64 ];
    static Char name0[ 64 + MAX_FILENAME ];
    static Char strippedDestFileName[ MAX_FILENAME ];
    Char*  gz;
    Char*  competitor;

    StrCopy( strippedDestFileName, destFileName );

    if ( ! StrCompare( strippedDestFileName + StrLen( strippedDestFileName ) - 3, ".gz" ) ) {
        strippedDestFileName[ StrLen( strippedDestFileName ) - 3 ] = 0;
    }

    if ( 0 <= compressLevel )
        gz = ".gz";
    else
        gz = "";

    if ( errNone != ZVFSOpenRead( srcVol, srcPath, &src, CRYPT( srcVol ) ) ) {
        Message( ERROR, "...error opening" );
        Pause( 1 );
        errorInBackup = true;
        return;
    }

    mdate = src.date;

    if ( errNone != ZVFSFileSize( &src, &size ) ) {
        Message( ERROR, "...error in size checking" );
        ZVFSFileClose( &src, false );
        Pause( 1 );
        errorInBackup = true;
        return;
    }

    if ( size == 0 ) {
        Message( ERROR, "\b...empty" );
        ZVFSFileClose( &src, false );
        return;
    }

    competitor = NULL;

    StrPrintF( destName, "%s/%s%s", destDir, strippedDestFileName, gz );

    if ( ! doAll ) {
        for ( i = 0 ; i < countOutDir ; i++ ) {
             if ( SAME_BASE( destFileName, namesOutDir[ i ] ) ) {

                 competitor = namesOutDir[ i ];

                 StrPrintF( name0, "%s/%s", destDir, competitor );

//                 Message( name0 );

                 if ( SameFile( destVol, name0, &src, size ) ) {
                     Message( INFO, "\b...unchanged" );
                     return;
                 }
                 else {
                     if ( errNone != ZVFSOpenRead( srcVol, srcPath, &src, CRYPT( srcVol ) ) ) {
                          Message( ERROR, "...error opening" );
                          Pause( 1 );
                          errorInBackup = true;
                          return;
                     }
                     break;
                 }
             }
        }
    }

    if ( useTemp ) {
        StrPrintF( destName, "%s/%s%s", destDir, tempName, gz );
    }

    FileDeleteBoth( destVol, destName );

    if ( ! doAll ) {
        EnsureSpace( destVol, destDir, size, srcIndex );
    }

    err = VFSFileCreate( destVol, destName );

    if ( err != errNone ) {
        ZVFSFileClose( &src, false );
        VFSFileDelete( destVol, destName );
        Message( ERROR, "...error creating" );
        errorInBackup = true;
        Pause( 1 );
        return;
    }

    if ( errNone != ZVFSOpenWrite( destVol, destName, &dest, compressLevel, mdate, CRYPT( destVol ) ) ) {
        ZVFSFileClose( &src, false );
        Message( ERROR, "...error opening output" );
        VFSFileDelete( destVol, destName );
        errorInBackup = true;
        Pause( 1 );
        return;
    }

    VFSFileSetAttributes( dest.ref, 0x20 );

    success = CopyFileRef( &dest, &src, size );

#ifdef RANDOM_FAILURE
    if ( SysRandom(0)%10 == 0 ) {
        Message( ERROR, "\bRandom failure" );
        success = false;
    }
#endif

    err = ZVFSFileClose( &dest, false );

    if ( err != errNone ) {
        Message( ERROR, "...error in outputing" );
        success = false;
    }

    err = ZVFSFileClose( &src, true );

    if ( err != errNone ) {
        Message( ERROR, "...error in input file" );
        success = false;
    }

    if ( useTemp && success ) {

        StrPrintF( name0, "%s/%s", destDir, strippedDestFileName );

        FileDeleteBoth( destVol, name0 );

        StrPrintF( name0, "%s%s", strippedDestFileName, gz );

        success = ( errNone == VFSFileRename( destVol, destName, name0 ) );

        if ( ! success ) {
            Message( ERROR, "...error renaming temporary" );
            success = false;
            Pause( 1 );
        }
    }

    if ( success ) {
        if ( volRefSrc == volRefHidden ) {
            retries[ countRetries ].verifyOnly = true;
            retries[ countRetries ].srcIndex   = srcIndex;
            retries[ countRetries ].crc        = dest.crc32;
            retries[ countRetries ].size       = dest.totalUncompressedSize;
            countRetries++;
        }
    }
    else {
        VFSFileDelete( destVol, destName );
        errorInBackup = true;
        retries[ countRetries ].verifyOnly = false;
        retries[ countRetries ].srcIndex   = srcIndex;
        countRetries++;
    }
}



Char** ScanDir( UInt16 vol, const Char* path, Int32* countP, ScanType type )
{
    UInt32        count;
    UInt32        iterator;
    UInt32        i;
    FileInfoType  info;
    FileRef       dirRef;
    Err           err;
    static Char          nameBuf[ 64 + MAX_FILENAME ];
    Char**        names;

    if ( vol == 0 ) {
        *countP = 0;
        return NULL;
    }

    *countP = -1;

    err = VFSFileOpen( vol, path, vfsModeRead, &dirRef );

    if ( err != errNone ) {
        return NULL;
    }

    info.nameP      = nameBuf;
    info.nameBufLen = MAX_FILENAME;


    count    = 0;
    iterator = vfsIteratorStart;

    while ( iterator != vfsIteratorStop  ) {
        err = VFSDirEntryEnumerate( dirRef, &iterator, &info );

        if ( err != errNone )
            break;

        if ( ( type == DIRS && !( info.attributes & vfsFileAttrDirectory ) ) ||
             ( type == BACKUP_FILES && ( info.attributes & vfsFileAttrDirectory ) ) ||
             ( type == BACKUP_FILES && ! ValidBackupName( info.nameP ) ) )
            continue;

        count++;
    }

    if ( count == 0 )
        names = NULL;
    else {
        names = MemPtrNewBig( count * sizeof( Char* ) );
        if ( names == NULL ) {
            Message( FATALERROR, "Out of memory" );
            VFSFileClose( dirRef );
            return NULL;
        }
    }

    i     = 0;
    iterator = vfsIteratorStart;

    while ( iterator != vfsIteratorStop && i < count ) {
        err = VFSDirEntryEnumerate( dirRef, &iterator, &info );

        if ( err != errNone )
            break;

        if ( ( type == DIRS && !( info.attributes & vfsFileAttrDirectory ) ) ||
             ( type == BACKUP_FILES && ( info.attributes & vfsFileAttrDirectory ) ) ||
             ( type == BACKUP_FILES && ! ValidBackupName( info.nameP ) ) )
            continue;

        names[ i ] = MemPtrNew( StrLen( info.nameP ) + 1 );

        if ( names[ i ] == NULL )
            continue;

        StrCopy( names[ i ], info.nameP );

        i++;
    }

    VFSFileClose( dirRef );

    *countP = i;

    return names;
}


void DeleteUniqueFiles( UInt16 vol, Char* dir, Char** names, Int32 count, Char** compareAgainst, Int32 compareCount )
{
    static Char          nameBuf[ 64 + MAX_FILENAME ];
    Int32         i;

    for ( i = 0 ; i < count ; i++ ) {
         Int32 j;

         ProgressBar( THICK, i, count );

         for ( j = 0 ; j < compareCount ; j++ ) {
              if ( names[ i ][ 0 ] == compareAgainst[ j ][ 0 ] && SameBaseFileName( names[ i ], compareAgainst[ j ] ) )
                  break;
         }

         if ( j < compareCount )
             continue;

         StrPrintF( nameBuf, "%s/%s", dir, names[ i ] );
         VFSFileDelete( vol, nameBuf );
    }
}



UInt16 FindVolume( VolumeType type, VolumeInfoType* infoP )
{
    UInt32        iterator;
    UInt16        vol;
    VolumeInfoType info;

    if ( type == CARD ) {
        /* Prefer a real CARD/MMC card if possible. */
        vol = FindVolume( pref.internalDrive ? INTERNAL_DRIVE : STRICT_SD, infoP );
        if ( vol != 0 )
            return vol;
        if ( pref.internalDrive )
            type = NON_REMOVABLE;
    }

    if ( infoP == NULL )
        infoP = &info;

    if ( type == HIDDEN )
        iterator = vfsIteratorStart | vfsIncludePrivateVolumes;
    else
        iterator = vfsIteratorStart;

    while ( iterator != vfsIteratorStop ) {

       if ( errNone == VFSVolumeEnumerate( &vol, &iterator ) &&
            errNone == VFSVolumeInfo( vol, infoP ) &&
            ( ( type == HIDDEN && ( ( infoP->attributes & vfsVolumeAttrHidden )
                               || infoP->mediaType == 'NFFS' ) ) ||
              ( type == STRICT_SD && ! ( infoP->attributes & vfsVolumeAttrNonRemovable ) &&
                IS_SD( infoP->mediaType ) ) ||
              ( type == NON_REMOVABLE && ( infoP->attributes & vfsVolumeAttrNonRemovable ) ) ||
              ( type == CARD && ! ( infoP->attributes & vfsVolumeAttrNonRemovable ) ) ||
              ( type == INTERNAL_DRIVE && IS_INTERNAL_DRIVE( infoP->mediaType ) &&
                ( infoP->attributes & vfsVolumeAttrNonRemovable ) )
              )
           ) {
           return vol;
       }
    }

    return 0;
}



Boolean SaveComment( UInt16 vol, Char* dir, Char* comment )
{
    static Char buf[ 64 + MAX_FILENAME ];
    FileRef ref;
    Err     err;

    StrPrintF( buf, "%s/%s", dir, oldCommentName );
    VFSFileDelete( vol, buf );
    StrPrintF( buf, "%s/%s", dir, commentName );
    VFSFileDelete( vol, buf );
    VFSFileCreate( vol, buf );

    if ( errNone != VFSFileOpen( vol, buf, vfsModeWrite, &ref ) )
        return false;

    err = VFSFileWrite( ref, StrLen( comment ), comment, NULL );

    VFSFileClose( ref );

    return err == errNone;
}



Boolean SaveCrypt( UInt16 vol, Char* dir, UInt32 cryptID )
{
    static Char buf[ 64 + MAX_FILENAME ];
    FileRef ref;
    Err     err;

    StrPrintF( buf, "%s/%s", dir, oldCryptName );
    VFSFileDelete( vol, buf );
    StrPrintF( buf, "%s/%s", dir, cryptName );
    VFSFileDelete( vol, buf );

    if ( cryptID == 0 ) {
        /* TODO: error check? */
        if ( errNone == VFSFileOpen( vol, buf, vfsModeRead, &ref ) ) {
            VFSFileClose( ref );

            return false;
        }
        return true;
    }

    VFSFileCreate( vol, buf );

    if ( errNone != VFSFileOpen( vol, buf, vfsModeWrite, &ref ) )
        return false;

    err = VFSFileWrite( ref, sizeof( UInt32 ), &cryptID, NULL );

    VFSFileClose( ref );

    return err == errNone;
}



UInt32 GetCrypt( UInt16 vol, Char* dir )
{
    static Char buf[ 64 + MAX_FILENAME ];
    FileRef ref;
    UInt32  cryptID;
    Err     err;

    StrPrintF( buf, "%s/%s", dir, cryptName );

    if ( errNone != VFSFileOpen( vol, buf, vfsModeRead, &ref ) ) {
        StrPrintF( buf, "%s/%s", dir, oldCryptName );
        if ( errNone != VFSFileOpen( vol, buf, vfsModeRead, &ref ) )
            return 0;
    }

    err = VFSFileRead( ref, sizeof( UInt32 ), &cryptID, NULL );

    VFSFileClose( ref );

    return ( err == errNone ) ? cryptID : 0xFFFFFFFFul;

}



Boolean MakeDir( UInt16 vol, const Char* dir )
{
    Err err;

    err = VFSDirCreate( vol, dir );

    return err == errNone || err == vfsErrFileAlreadyExists;
}


/* 0xFF if not hex */
UInt8 GetHexNibble( Char c )
{
    if ( '0' <= c && c <= '9' )
        return c - '0';
    if ( 'a' <= c && c <= 'f' )
        return c - 'a' + 0xA;
    if ( 'A' <= c && c <= 'F' )
        return c - 'A' + 0xA;
    return 0xFF;
}


UInt8 GetHexByte( Char* s )
{
    return GetHexNibble( s[ 0 ] ) * 16 + GetHexNibble( s[ 1 ] );
}


static Boolean IsHexString( Char* s, UInt16 n )
{
    while ( n-- ) {
        if ( 0xFF == GetHexNibble( *s++ ) )
            return false;
    }
    
    return true;
}



static Boolean IsStandardFileName( Char* s )
{
    UInt16 len;

    len = StrLen( s );

    return 9 < len && s[ len - 9 ] == '_' && IsHexString( s + len - 8, 8 );
}


void FixName( Char* out, Char* in )
{
    Char* out0;
    UInt16 len;

    out0 = out;

    while ( *in ) {
        if ( *in != '%' ) {
            *out++ = *in++;
        }
        else {
            if ( in[ 1 ] && in[ 2 ] ) {
                *out++ = GetHexByte( in + 1 );
                in += 3;

            }
        }
    }

    *out = 0;
    
    len = StrLen( out0 );

    if ( 3 < len && ! StrCompare( out0 + len - 3, ".gz" ) ) {
        out0[ StrLen( out0 ) - 3 ] = 0;
        len = StrLen( out0 );
    }

    if ( IsStandardFileName( out0 ) )
        out0[ len - 9 ] = 0;
    else if ( 1 < len && out0[ len - 1 ] == '_' )
        out0[ len - 1 ] = 0;
}



Boolean ValidBackupName( Char* name )
{
    UInt16 len;

    if ( name[ 0 ] == '%' && name[ 1 ] == '%' )
        return false;
        
    if ( ! StrNCompare( name, badBackupFileName, StrLen( badBackupFileName ) ) )
        return false;

//if ( name[ 0 ] != 'U' || name[ 1 ] != 'n' )
//return false;

    len = StrLen( name );

    if ( 3 < len && ! StrCaselessCompare( name + len - 3, ".gz" ) )
        len -= 3;

    if ( 1 < len && name[ len - 1 ] == '_' )
        return true;

    return 9 < len && name[ len - 9 ] == '_' && IsHexString( name + len - 8, 8 );
}




Boolean FindMe( UInt16* card, LocalID* id )
{
    DmSearchStateType state;

    return errNone == DmGetNextDatabaseByTypeCreator( true, &state,
               'appl', appCreator,
               false, card, id );
}



Boolean CopyMeTo( UInt16 volRef, const Char* path )
{
    UInt16 card;
    LocalID id;
    
    if ( ! FindMe( &card, &id ) )
        return false;
        
    return errNone == VFSExportDatabaseToFile( volRef, path, card, id );
}



#ifdef DIM
void TempLCDBrightness( UInt8 brightness )
{
    UInt8 data[ 4 ];
    UInt16 size;

    size = 0;

    if ( noPreferenceFound ==
             PrefGetAppPreferences( 'P050', 0, NULL, &size, false ) ||
         size != 4 ||
         noPreferenceFound ==
             PrefGetAppPreferences( 'P050', 0, data, &size, false )
         ) {
        return;
    }

    oldBrightness = SysLCDBrightness( false, 0 );
    SysLCDBrightness( true, brightness );

    PrefSetAppPreferences( 'P050', 0, 0, data,
        4, false );
}
#endif



void CleanupBackup( void )
{
    Int32 i;

    for ( i = 0 ; i < countOutDir ; i++ ) {
        SafeMemPtrFree( namesOutDir[ i ] );
    }

    countOutDir = 0;
    SafeMemPtrFree( namesOutDir );
    namesOutDir = NULL;

    if ( disabledNotifications ) {
        for ( i = 0 ; i < numNotifications ; i++ )
             Unregister( notifications[ i ], -128 );

        AlmEnableNotification( true );

        disabledNotifications = false;
    }

    for ( i = 0 ; i < countInDir ; i++ ) {
        SafeMemPtrFree( namesInDir[ i ] );
    }

    countInDir = 0;
    SafeMemPtrFree( namesInDir );
    namesInDir = NULL;

    SafeMemPtrFree( retries );
    countRetries = 0;
    retries      = NULL;

    CloseLog();

#ifdef DIM
    if ( 0 <= oldBrightness ) {
        SysLCDBrightness( true, oldBrightness );
        oldBrightness = -1;
    }
#endif    
}


void CloseCrypt( void )
{
    if ( crypt.Close != NULL ) {
        crypt.Close();
        crypt.Crypt = NULL;
        crypt.Close = NULL;
    }
}


void FatalBackupError( void )
{
    if ( nvfsOff ) {
        PauseReset();
    }
    else {
        CleanupBackup();
        if ( scheduled ) {
            EvtEnqueueKey( vchrPowerOff, 0, commandKeyMask );
        }
        else {
            EvtResetAutoOffTimer();
        }
        FrmAlert( altBackupFatalError );
        FtrUnregister( appCreator, FTR_BUSY );
        CloseCrypt();
    }
}


#ifdef FLUSH
void DBCacheInfo( void )
{
    UInt32 total;
    UInt32 free;
    UInt32 chunk;
    static Char   message[ 80 ];

    total = StorageSize( true, &free, &chunk );

    StrPrintF( message, "DBCache total: %ldK, Free: %ld%%, Max chunk: %ldK", total/1024, 100 * free/total, chunk/1024 );

    Message( INFO, message );
}
#endif



UInt32 DoCRC32( UInt16 vol, const Char* path, UInt32* sizeP )
{
    ZFile  zref;
    UInt32 crc = 0;
    UInt32 left;

    *sizeP = 0;

    if ( errNone != ZVFSOpenRead( vol, path, &zref, CRYPT( vol ) ) )
        return 0;

    if ( errNone != ZVFSFileSize( &zref, sizeP ) ) {
        ZVFSFileClose( &zref, false );

        return 0;
    }

    left = *sizeP;

    while ( 0 < left ) {
        UInt16 size1;
        UInt16 didRead;

        size1 = BUFFER_SIZE;

        if ( left < size1 )
            size1 = left;

        if ( errNone != ZVFSRead( &zref, size1, buffer1, &didRead ) ||
             didRead < size1 ) {
            ZVFSFileClose( &zref, false );
            *sizeP = 0;
            return 0;
        }

        crc = CRC32( crc, buffer1, didRead );

        left -= size1;
    }

    ZVFSFileClose( &zref, false );

    return crc;
}




static Boolean GetPassword( Char* password )
{
    switch ( FrmCustomResponseAlert( altKey, "", "", "",
                 password, MAX_PASSWORD_SIZE, KeyVerify ) ) {
        case 0:
            ClearScreen();
            return true;
        default:
            ClearScreen();
            return false;
    }
}



static Boolean HandleCrypt( VolumeType srcType, Char* destDir, Char* srcDir )
{
    static Char password[ MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE ];
    Char*  name;
    UInt16 curCardVol;

    curCardVol     = FindVolume( CARD, NULL );

    if ( curCardVol == 0 )
        return false;

    if ( srcType == HIDDEN ) {

        if ( ! pref.doCrypt ) {
            CryptSetup( &crypt, 0, NULL );
            SaveCrypt( curCardVol, destDir, 0 );
            return true;
        }

        if ( pref.haveKey ) {
            PasswordDecrypt( password, pref.key );
        }
        else {
            if ( ! GetPassword( password ) )
                return false;
        }

        name = CryptSetup( &crypt, DEFAULT_ENCRYPTION, password );
        MemZero( password, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE );

        if ( name == NULL )
            return false;

        SaveCrypt( curCardVol, destDir, DEFAULT_ENCRYPTION );
    }
    else {
        UInt32 cryptID;

        *password = 0;

        cryptID = GetCrypt( curCardVol, srcDir );

        if ( cryptID == 0xFFFFFFFFul )
            return false;

        if ( cryptID != 0 ) {
            if ( pref.haveKey ) {
                PasswordDecrypt( password, pref.key );
            }
            else {
                if ( ! GetPassword( password ) )
                    return false;
            }
        }

        name = CryptSetup( &crypt, cryptID, password );
        MemZero( password, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE );

        if ( name == NULL )
            return false;
    }

    Message( INFO, "Encryption:" );
    Message( INFO, name );
    
    return true;
}



Boolean CheckBattery( void )
{
    UInt16        threshold;
    UInt16        voltage;
    Boolean         plugged;

    voltage = SysBatteryInfo( false, &threshold, NULL, NULL, NULL, &plugged, NULL );

    return plugged || threshold * ( UInt32 )THRESHOLD_1000 / 1000 <= voltage;
}



Boolean DoBackup( Char* destDir, VolumeType destType, Char* srcDir, VolumeType srcType, Boolean doAll, Boolean partial,
               Char* comment, Char* renameTo, Int8 curCompressLevel,
               Char** selectFiles, Int32 numSelectFiles, Boolean* selection )
{
    VFSSlotMountParamType param;
    Int32         i;
    UInt16        j;
    UInt32        depth;
    Boolean       nvfsOff;
#ifdef FLUSH
    Boolean       doFlush;
#endif
    static Char     inName[ MAX_FILENAME + 64 ];
    static Char     temp[ MAX_FILENAME ];
    
    if ( ! CheckBattery() ) {
        Message( FATALERROR, "Insufficient battery." );
        FatalBackupError();
        return false;
    }
    
    locked = ( 0 != PrefGetPreference( prefDeviceLocked ) );

    FtrSet( appCreator, FTR_BUSY, 1 );

#ifdef DIM
    oldBrightness = -1;
#endif

    if ( selectFiles != NULL ) {
        if ( numSelectFiles < 0 ) {
            FatalBackupError();
            return false;
        }

        partial = true;
        doAll   = false;
    }

    if ( destType != HIDDEN )
        compressLevel = curCompressLevel;
    else
        compressLevel = -1;

//    depth = 16;
//    WinScreenMode( winScreenModeSet, NULL, NULL, &depth, NULL );

    ClearScreen();
    LoadMessageFont();

    if ( destType == HIDDEN )
        Message( INFO, "From card to hidden..." );
    else
        Message( INFO, "From hidden to card..." );

    if ( ! HandleCrypt( srcType, destDir, srcDir ) ) {
        FrmAlert( altCryptError );
        CloseCrypt();
        return false;
    }

    Message( INFO, "Checking for card..." );
    volRefCard = FindVolume( CARD, &cardInfo );

    if ( 0 == volRefCard ) {
        FatalError( altNoCard );
    }

    buffer1 = SafeMemPtrNew( BUFFER_SIZE );
    buffer2 = SafeMemPtrNew( BUFFER_SIZE );

//    if ( destType == CARD ) {
        InitLog( volRefCard, ( srcType == HIDDEN ) ? destDir : srcDir );
        if ( logBuffer != NULL )
            Message( INFO, "Initialized log..." );
//    }

    if ( destType == CARD && selectFiles == NULL && pref.copySelf ) {
        Char* path;

        Message( INFO, "Copying NVBackup to card..." );

        if ( pref.internalDrive ) {
            MakeDir( volRefCard, "/Applications" );
            path = "/Applications/NVBackup.prc";
        }
        else {
            MakeDir( volRefCard, "/PALM" );
            MakeDir( volRefCard, "/PALM/Launcher" );
            path = "/PALM/Launcher/NVBackup.prc";
        }

        VFSFileDelete( volRefCard, path );
        if ( ! CopyMeTo( volRefCard, path ) )
            Message( ERROR, "...error copying NVBackup" );
    }

//    if ( destType == HIDDEN && selectFiles == NULL ) {
//        Message( INFO, "Trying to disable hacks..." );
//        DisableHacks();
//    }

    Message( INFO, "Locating hidden volume..." );
    volRefHidden = FindVolume( HIDDEN, &hiddenInfo );

    if ( 0 == volRefHidden ) {
        CloseLog();
        FatalError( altNoHidden );
    }

//if ( srcType == HIDDEN ) {
//VFSFileRename( volRefHidden, "PALM_DM", "PALM_D1" );
//srcDir = "/PALM_D1";
//}

#ifdef FLUSH
    doFlush = false;

    if ( srcType == HIDDEN ) {
        switch ( pref.flush ) {
            case FLUSH_ALWAYS:
                doFlush = true;
                break;
            case FLUSH_NEVER:
                doFlush = false;
                break;
            case FLUSH_AUTO: {
                UInt32 total;
                UInt32 free;
                UInt32 chunk;

                Message( INFO, "Checking dbCache size..." );

                DBCacheInfo();

                total = StorageSize( true, &free, &chunk );

                if ( free < total * DBCACHE_FREE_PERCENTAGE / 100 ||
                     chunk < total * DBCACHE_CHUNK_PERCENTAGE / 100 )
                    doFlush = true;
            }
        }
    }

    if ( doFlush ) {
        if ( pref.flush != FLUSH_AUTO )
            DBCacheInfo();

        Message( INFO, "Flushing dbCache..." );

        DBCacheFlush();

        DBCacheInfo();
    }
#endif

    Message( INFO, "Flushing cached writes..." );
    DmSync();

    Message( INFO, "Disabling alarms and notifications..." );
    AlmEnableNotification( false );

    if ( destType != HIDDEN ) {
        /* Disable fewer notifications when backing up than when restoring. */
        notifications = fewerNotifications;
        numNotifications = sizeof( fewerNotifications ) / sizeof( *fewerNotifications );
    }
    else {
        notifications = moreNotifications;
        numNotifications = sizeof( moreNotifications ) / sizeof( *moreNotifications );
    }

    for ( i = 0 ; i < numNotifications ; i++ )
         Register( notifications[ i ], DisableNotification, -128 );

    disabledNotifications = true;

    if ( destType == HIDDEN && ! pref.testMode ) {
        nvfsOff = true;

        Message( INFO, "Unmounting hidden volume..." );
        if ( errNone != VFSVolumeUnmount( volRefHidden ) ) {
            Message( FATALERROR, "Error unmounting." );
            PauseReset();
        }

        Message( INFO, "Remounting hidden volume..." );
        param.vfsMountParam.mountClass = vfsMountClass_SlotDriver;
        param.slotLibRefNum            = hiddenInfo.slotLibRefNum;
        param.slotRefNum               = hiddenInfo.slotRefNum;

        if ( errNone != VFSVolumeMount( 0, 0, ( VFSAnyMountParamType* )&param ) ) {
            Message( FATALERROR, "Error mounting." );
            PauseReset();
        }

        volRefHidden = param.vfsMountParam.volRefNum;
    }
    else {
        nvfsOff = false;
        Message( INFO, "(No need to disable NVFS.  Thanks, Dmitry!)" );
#ifdef DIM
        if ( scheduled && pref.dimOnScheduled && destType == CARD ) {
            TempLCDBrightness( 0 );
        }
#endif
    }
    
    if ( destType == HIDDEN ) {
        volRefDest = volRefHidden;
        Message( INFO, "Locating source volume..." );
        volRefSrc  = FindVolume( srcType, &ignoreInfo );
        if ( volRefSrc == 0 ) {
            Message( ERROR, "Error locating source." );
            FatalBackupError();
            return false;
        }
    }
    else {
        volRefSrc = volRefHidden;
        Message( INFO, "Locating destination volume..." );
        volRefDest  = FindVolume( destType, &ignoreInfo );
        if ( volRefDest == 0 ) {
            Message( FATALERROR, "Error locating destination." );
            FatalBackupError();
            return false;
        }
    }

    if ( pref.testMode ) {
        volRefDest = volRefCard;
        volRefSrc  = volRefCard;
    }

    if ( destType == CARD ) {
        Message( INFO, "Saving comment..." );
        if ( ! SaveComment( volRefDest, destDir, comment ) ) {
            Message( ERROR, "Error saving comment." );
        }
    }
    
    Message( INFO, "Scanning destination..." );
    namesOutDir = ScanDir( volRefDest, destDir, &countOutDir, BACKUP_FILES );

    if ( countOutDir < 0 ) {
        Message( FATALERROR, "Error scanning output directory." );
        FatalBackupError();
        return false;
    }

    Message( INFO, "Scanning source..." );
    namesInDir = ScanDir( volRefSrc, srcDir, &countInDir, BACKUP_FILES );

    if ( countInDir < 0 ) {
        Message( FATALERROR, "Error scanning directory." );
        FatalBackupError();
        return false;
    }

    if ( 0 < countInDir ) {
        retries = MemPtrNewBig( sizeof( RetryType ) * countInDir );
        if ( retries == NULL ) {
            Message( FATALERROR, "Error allocating memory." );
            FatalBackupError();
            return false;
        }
        countRetries = 0;
    }

    if ( ! CheckBattery() ) {
        Message( FATALERROR, "Insufficient battery." );
        FatalBackupError();
        return false;
    }

    if ( ! partial ) {
        Message( INFO, "Cleaning out unneeded files..." );
        DeleteUniqueFiles( volRefDest, destDir, namesOutDir, countOutDir, doAll ? NULL : namesInDir,
                           doAll ? 0 : countInDir );
    }

    Message( INFO, "Copying data from" );
    Message( INFO, srcDir );
    Message( INFO, "to" );
    Message( INFO, destDir );

    for ( i = 0 ; i < countInDir ; i++ ) {
        static Boolean skip;

        ProgressBar( 0, i, countInDir );

        if ( namesInDir[ i ] == NULL )
            continue;

        skip = false;

        if ( selectFiles != NULL ) {
            Int32  j;

            skip = true;

            for ( j = 0 ; j < numSelectFiles ; j++ ) {
                 if ( SAME_BASE( namesInDir[ i ], selectFiles[ j ] ) ) {
                     skip = ! selection[ j ];
                     break;
                 }
            }
        }

        if ( skip )
            continue;

        StrPrintF( inName, "%s/%s", srcDir, namesInDir[ i ] );

        FixName( temp, namesInDir[ i ] );
        Message( INFO, temp );
        ProgressBar( 1, 0, 1 );

        CopyFile( volRefDest, destDir, namesInDir[ i ], volRefSrc, inName, i,
                  doAll,
                  true /* destType == CARD */ );
    }

    for ( j = 0 ; j < MAX_RETRY_COUNT && 0 < countRetries ; j++ ) {
        Int32 oldCountRetries;
        Char  s[ 15 ];

        StrPrintF( s, "Pass %d...", j + 2 );
        Message( INFO, s );

        errorInBackup   = false;
        oldCountRetries = countRetries;
        countRetries    = 0;

        for ( i = 0 ; i < oldCountRetries ; i++ ) {
             Int32   index;                    
             UInt32  size;

             index = retries[ i ].srcIndex;

             ProgressBar( 0, i, oldCountRetries );

             StrPrintF( inName, "%s/%s", srcDir, namesInDir[ index ] );

             if ( ! retries[ i ].verifyOnly ||
                  DoCRC32( volRefSrc, inName, &size ) != retries[ i ].crc ||
                  size != retries[ i ].size ) {
                 FixName( temp, namesInDir[ index ] );

                 Message( INFO, temp );
                 ProgressBar( 1, 0, 1 );

                 CopyFile( volRefDest, destDir, namesInDir[ index ], volRefSrc, inName, index,
                           true /* doAll */,
                           true /* destType == CARD */ );
             }
        }
    }

    if ( errorInBackup ) {
        Message( ERROR, "There was at least one error." );
        CleanupBackup();
        if ( ! nvfsOff ) {
            if ( scheduled )
                EvtEnqueueKey( vchrPowerOff, 0, commandKeyMask );
            else
                EvtResetAutoOffTimer();
            FrmAlert( altIncompleteBackup );
            FtrUnregister( appCreator, FTR_BUSY );
            CloseCrypt();
            return false;
        }
    }
    else {
        if ( 0 < countRetries )
            Message( IMPORTANT, "Could not verify all files." );
        else
            Message( IMPORTANT, "Copied with no errors." );
    }

    if ( ! errorInBackup && renameTo != NULL ) {
        Message( INFO, "Renaming output directory..." );
        VFSFileRename( volRefDest, destDir, renameTo );
    }

    ProgressBar( 0, 1, 1 );

    Message( IMPORTANT, "Done!" );

    if ( ! errorInBackup )
        CleanupBackup();

    if ( nvfsOff ) {
        PauseReset();
    }
    else {
        Debug( "<<ClearScreen()>>" );
        ClearScreen();
#if 0
        while ( EvtEventAvail() ) {
            EventType e;

            EvtGetEvent( &e, 0 );
        }
#endif
        EvtFlushKeyQueue();
        EvtFlushPenQueue();
        Debug( "<<cleared events>>" );
    }

    EvtResetAutoOffTimer();
    FtrUnregister( appCreator, FTR_BUSY );
    Debug( "<<returning>>" );
    CloseCrypt();
    return ! errorInBackup;
}



