/* Id: UI.c,v 1.40 2006/11/05 14:49:04 arpruss Exp $ */
#include <PalmOS.h>
#include "Backup.h"
#include "DIA.h"
#include "ftp.h"

FTP_SECTION Boolean SNTP( Char* h );

#include "Backup_Res.h"

#define SNOOZE_TIME ( 60ul * 10 ) // 10 minutes

#define DATE_LENGTH 14 // YYYYMMDDHHMMSS
#define MAX_FRIENDLY 80

#define CHECKMARK_SIZE 16

#define OFFSET( type, field ) ( ( UInt32 )( ( UInt8* )&( ( ( type * )NULL ) ->field ) - ( UInt8* )NULL ) )
#define PREF_OFFSET( field )  OFFSET( PreferenceType, field )
#define BOOLEAN_PTR( var, offset ) ( ( Boolean* )( ( UInt8* )( var ) + ( offset ) ) )
#define PREF_BOOLEAN_PTR( offset ) BOOLEAN_PTR( &pref, offset )

#define DO_ACTION firstUserEvent

typedef enum {
     selectRestore = 0,
     selectBackup
} SelectType;

static void Init_frmMain( void );
static void Restore( Int16 sel, Boolean lowMemory );
static void SelectGoRestore( void );
static void SelectGoBackup( void );
static void SelectRange( Boolean select );

typedef void (*SelectGoFunction)( void );

typedef struct {
     const Char*      label;
     SelectGoFunction Action;
     VolumeType       volumeSourceType;
} SelectGoType;

SelectGoType selectGo[] =
{
 { "Restore", SelectGoRestore, CARD },
 { "Backup",  SelectGoBackup,  HIDDEN }
};

typedef struct {
    UInt16   id;
    UInt32   offset;
    Boolean  reverse;
    Boolean  defaultValue; // this is the value in the prefs, not the display value
    Boolean  advanced;
} BooleanOptionType;

BooleanOptionType booleanOptions[] =
{
    { chkResetAfterScheduled, PREF_OFFSET( resetAfterScheduled ), false, false, true },
    { chkCopySelf, PREF_OFFSET( copySelf ), false, true, false },
//    { chkDimOnScheduled, PREF_OFFSET( dimOnScheduled ), false, false, false },
    { chkTrustDates, PREF_OFFSET( trustDates ), false, false, true },
    { chkSetTime, PREF_OFFSET( setTime ), false, false, true },
//    { chkDisableFewer, PREF_OFFSET( disableFewer ), false, false, true }
};

#define NUM_BOOLEAN_OPTIONS ( sizeof booleanOptions / sizeof *booleanOptions )

Char    comment[ MAX_COMMENT_SIZE + 1 ];
ActionType afterComment = NOOP;
ActionType nextAction   = NOOP;
Int32      afterCommentArgument;
Int32      nextActionArgument;

static Char* selectSourceDir;
static SelectType select;
static Int32    selectSet;
UInt16   curCardVol;
static SetType* sets  = NULL;
static Char**   friendlyNames = NULL;
static Int32    setCount = 0;

static Char**   fileList = NULL;
static Boolean* fileSelection = NULL;
static Int32    fileCount = 0;

static Boolean  terminate = false;
Boolean  redoList = false;
static Boolean  redrawList = false;
Boolean  scheduled = false;
PreferenceType pref;

void SavePrefs0( PreferenceType* p );
static void DoAction( void );
static void Backup( Int16 sel );
static void ErrorInBackup( void );
HANDLER_SECTION static Boolean Handler_frmMain( EventType* event );
HANDLER_SECTION static Boolean Handler_frmAbout( EventType* event );
HANDLER_SECTION static Boolean Handler_frmPrefs( EventType* event );
HANDLER_SECTION static Boolean Handler_frmAdvanced( EventType* event );
HANDLER_SECTION static Boolean Handler_frmSelect( EventType* event );
MISC_SECTION static Boolean Warnings( void );

Boolean FindBadBackupFile( UInt16* card, LocalID* id )
{
    DmSearchStateType state;

    return errNone == DmGetNextDatabaseByTypeCreator( true, &state,
               badBackupFileType, appCreator,
               false, card, id );
}



void FatalError( UInt16 alert )
{
     if ( scheduled )
         SysReset();
     FrmAlert( alert );
     ErrThrow( 1 );
}



#if 0
void LockIfNeeded( PreferenceType* p )
{
     if ( p->needLock ) {
         p->needLock = false;
         SavePrefs0( p );
         if ( PwdExists() )
             EvtEnqueueKey( vchrLock, 0, commandKeyMask );
     }
}
#endif



Boolean Confirm( UInt16 alert )
{
     return 0 == FrmAlert( alert );
}


void* SafeMemPtrNew( UInt32 size )
{
    void* p;

    p = MemPtrNew( size );

    if ( p == NULL ) {
        FatalError( altNoMem );
    }

    return p;
}


void SafeMemPtrFree( void* p )
{
    if ( p != NULL )
        MemPtrFree( p );
}


static Boolean MakeDirs( UInt16 curCardVol )
{
    return MakeDir( curCardVol, "/PALM" ) &&
           MakeDir( curCardVol, "/PALM/Programs" ) &&
           MakeDir( curCardVol, backupDir );
}



static Boolean GetCard( void )
{
    curCardVol = FindVolume( CARD, NULL );

    if ( curCardVol == 0 ) {
        if ( ! pref.internalDrive ) {
            pref.internalDrive = true;
            
            curCardVol = FindVolume( CARD, NULL );

            if ( 0 != curCardVol && 0 == FrmAlert( altUseInternal ) ) {
                SavePrefs();
                return true;
            }

            pref.internalDrive = false;
            SavePrefs();

            return false;
        }
        else {
            pref.internalDrive = false;
            SavePrefs();

            curCardVol = FindVolume( CARD, NULL );

            if ( curCardVol == 0 ) {
                FrmAlert( altNoCard );
                return false;
            }
        }
    }

    MakeDirs( curCardVol );
    
    return true;
}


UInt16 GetObjectIndex( UInt16 id )
{
    return FrmGetObjectIndex( FrmGetActiveForm(), id );
}


void* GetObjectPtr( UInt16 id )
{
    return FrmGetObjectPtr( FrmGetActiveForm(), GetObjectIndex( id ) );
}


Int16 GetSelection( UInt16 lstID )
{
    return LstGetSelection( GetObjectPtr( lstID ) );
}


void SetList( UInt16 trgID, UInt16 lstID, Int16 sel )
{
    LstSetSelection( GetObjectPtr( lstID ),
        sel );
    CtlSetLabel( GetObjectPtr( trgID ),
        LstGetSelectionText( GetObjectPtr( lstID ),
        sel ) );
}

void SetValue( UInt16 id, UInt16 value )
{
    CtlSetValue( GetObjectPtr( id ), value );
}


UInt16 GetValue( UInt16 id )
{
    return CtlGetValue( GetObjectPtr( id ) );
}


void SavePrefs0( PreferenceType* p )
{
    PrefSetAppPreferences( appCreator, PREF_ID_MAIN, 0, p,
        sizeof( PreferenceType ), true );
}


void SavePrefs( void )
{
    SavePrefs0( &pref );
}


void LoadPrefs0( PreferenceType* p )
{
    UInt16 size;

    MemSet( p, sizeof( *p ), 0 );

    p->cycle = 1;
    p->scheduleHour = 1;
    p->compression  = 1;
#ifdef FLUSH
    p->flush        = FLUSH_NEVER;//FLUSH_AUTO;
#endif
    p->server.port  = 21;

    p->copySelf     = true;
    p->dimOnScheduled = true;

    size = 0;

    if ( noPreferenceFound ==
             PrefGetAppPreferences( appCreator, PREF_ID_MAIN, NULL, &size, true ) ) {
        return;
    }

    if ( sizeof( *p ) < size ) {
        size = sizeof( *p );
    }

    PrefGetAppPreferences( appCreator, PREF_ID_MAIN, p, &size, true );

    if ( p->cycle < MIN_CYCLE )
        p->cycle = MIN_CYCLE;
    if ( MAX_CYCLE < p->cycle )
        p->cycle = MAX_CYCLE;
    p->slowUpdate = false;
    p->testMode   = false;
}



void LoadPrefs( void )
{
    LoadPrefs0( &pref );

    if ( pref.haveDevicePassword ) {
        /* No longer needed in 1.12 */
        pref.haveDevicePassword = 0;
        MemZero( pref.devicePassword, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE );
    }
}


/* delta is a time change */
void UpdateSchedule0( PreferenceType* p, Int32 delta )
{
    UInt16 card;
    LocalID id;

    if ( ! FindMe( &card, &id ) )
        return;
        
    if ( p->lastScheduledTime ) {
        p->lastScheduledTime += delta;
    }

    if ( p->forceNextAlarm ) {
        p->forceNextAlarm += delta;
        if ( 86400l < p->forceNextAlarm - TimGetSeconds() )
            p->forceNextAlarm = 86400l + TimGetSeconds();
        AlmSetAlarm( card, id, 0, p->forceNextAlarm, true );
    }
    else if ( ! p->daily ) {
        AlmSetAlarm( card, id, 0, 0, true );
    }
    else {
        UInt32 seconds;
        UInt32 newSeconds;
        UInt32 alSeconds;
        UInt32 ref;
        
        AlmSetAlarm( card, id, 0, 0, true );

        seconds = TimGetSeconds();

        newSeconds = seconds / 86400l * 86400l + 60l * ( p->scheduleMinute + 60l * p->scheduleHour );

        alSeconds  = AlmGetAlarm( card, id, &ref );

        if ( alSeconds != newSeconds ) {
            if ( newSeconds <= seconds + 2 )
                newSeconds += 86400l;

            if ( newSeconds - p->lastScheduledTime < 60 && 
                 p->lastScheduledTime - newSeconds < 60 )
                newSeconds += 86400l;

            AlmSetAlarm( card, id, 0, newSeconds, true );
        }
    }
}



void UpdateSchedule( void )
{
    UpdateSchedule0( &pref, 0 );
}



static void ClearListData( void )
{
    UInt16 i;

    for ( i = 0 ; i < setCount ; i++ ) {
         SafeMemPtrFree( sets[ i ].path );
         SafeMemPtrFree( sets[ i ].friendly );
    }
    SafeMemPtrFree( sets );
    sets = NULL;
    setCount = 0;
    SafeMemPtrFree( friendlyNames );
    friendlyNames = NULL;
}


static void ClearFileList( void )
{
    Int32 i;

    for ( i = 0 ; i < fileCount ; i++ ) {
         SafeMemPtrFree( fileList[ i ] );
    }
    SafeMemPtrFree( fileList );
    SafeMemPtrFree( fileSelection );
    fileList  = NULL;
    fileSelection = NULL;
    fileCount = 0;
}



static Int16 CompareUnfixedStrings( void* a, void* b, Int32 discard )
{
    Char f[ MAX_FILENAME ];
    Char g[ MAX_FILENAME ];

    FixName( f, *( Char** )a );
    FixName( g, *( Char** )b );

    return StrCompare( f, g );
}




static void DrawFile( Int16 itemNum, RectangleType* bounds, Char** itemsText )
{
    FontID oldFont;
    Char   fixed[ MAX_FILENAME ];

    oldFont = FntSetFont( symbol11Font );

    WinDrawChar( !! fileSelection[ itemNum ], bounds->topLeft.x, bounds->topLeft.y );

    FntSetFont( oldFont );

    FixName( fixed, itemsText[ itemNum ] );

    WinDrawChars( fixed, StrLen( fixed ), bounds->topLeft.x + CHECKMARK_SIZE, bounds->topLeft.y );
}



static void SetupFileList( void )
{
    ListType* list;

    ClearFileList();

    fileList = ScanDir( FindVolume( selectGo[ select ].volumeSourceType, NULL ),
                   selectSourceDir, &fileCount, BACKUP_FILES );

    if ( fileCount <= 0 )
        fileCount = 0;
    else {
        fileSelection = SafeMemPtrNew( fileCount * sizeof( Boolean ) );
        MemSet( fileSelection, fileCount * sizeof( Boolean ), 0 );
        SysInsertionSort( fileList, fileCount, sizeof( Char* ), CompareUnfixedStrings, 0 );
    }

    list = GetObjectPtr( lstFiles );

    LstSetListChoices( list, fileList, fileCount );
    LstSetSelection( list, -1 );
    LstSetDrawFunction( list, DrawFile );
    LstDrawList( list );
}



Int16 CompareStrings( void* a, void* b, Int32 discard )
{
    return StrCompare( *( Char** )a, *( Char** )b );
}


Boolean IsIncomplete( Char* s )
{
    return StrLen( incompleteMarker ) <= StrLen( s ) &&
           ! StrCompare( s + StrLen( s ) - StrLen( incompleteMarker ), incompleteMarker );
}


static Char* Duplicate( Char* s )
{
     Char* p;
     
     p = SafeMemPtrNew( 1 + StrLen( s ) );
     StrCopy( p, s );

     return p;
}



static Char* GetInteger( Char* s, Int16* xP, UInt16 digits )
{
     *xP = 0;
     
     while ( digits-- ) {
         if ( *s < '0' || '9' < *s )
             return NULL;
         *xP = *xP * 10 + *s - '0';
         s++;
     }
     
     return s;
}



static Boolean GetDate( Char* s, DateTimeType* dtP )
{
     s = GetInteger( s, &dtP->year, 4 );
     if ( s == NULL )
         return false;
     s = GetInteger( s, &dtP->month, 2 );
     if ( s == NULL )
         return false;
     s = GetInteger( s, &dtP->day, 2 );
     if ( s == NULL )
         return false;
     s = GetInteger( s, &dtP->hour, 2 );
     if ( s == NULL )
         return false;
     s = GetInteger( s, &dtP->minute, 2 );
     if ( s == NULL )
         return false;
     s = GetInteger( s, &dtP->second, 2 );
     if ( s == NULL )
         return false;
     return true;
}



static void GetTimeFormat( TimeFormatType* tfP )
{
     UInt16 index;
     LmLocaleType loc;

     OmGetSystemLocale( &loc );

     if ( errNone != LmLocaleToIndex( &loc, &index ) ||
          errNone != LmGetLocaleSetting( index, lmChoiceTimeFormat, tfP, sizeof( TimeFormatType ) ) ) {
         *tfP = tfColonAMPM;
     }
}



static void GetDateFormat( DateFormatType* tfP )
{
     UInt16 index;
     LmLocaleType loc;

     OmGetSystemLocale( &loc );

     if ( errNone != LmLocaleToIndex( &loc, &index ) ||
          errNone != LmGetLocaleSetting( index, lmChoiceDateFormat, tfP, sizeof( DateFormatType ) ) ) {
         *tfP = dfMDYLongWithComma;
     }
}



Char* GetFriendly( Char* lastPart, Boolean forceIncomplete )
{
     Char s[ MAX_FRIENDLY + 1 ];
     DateTimeType dt;
     DateType date;
     DateType curDate;
     UInt32   days;
     UInt32   curDays;
     TimeFormatType tf;

     if ( ! GetDate( lastPart, &dt ) ) {
         if ( forceIncomplete ) {
             StrPrintF( s, lastPart );
             StrCat( s, " " );
             StrCat( s, incompleteMarker );
             return Duplicate( s );
         }
         return Duplicate( lastPart );
     }

     date.month = dt.month;
     date.day   = dt.day;
     date.year  = dt.year - 1904;

     days = DateToDays( date );
     curDays = TimGetSeconds() / 86400l;
     DateDaysToDate( curDays, &curDate );

     if ( curDays == days ) {
         StrCopy( s, "Today" );
     }
     else if ( curDays - 1 == days ) {
         StrCopy( s, "Yesterday" );
     }
     else if ( curDays + 1 == days ) {
         StrCopy( s, "Tomorrow" );
     }
     else {
         DateToAscii( dt.month, dt.day, dt.year, dfMDYLongWithComma, s );
         if ( curDate.year + 1904 == dt.year ) {
             Char yearComp[ 12 ];

             StrPrintF( yearComp, ", %d", curDate.year + 1904 );

             if ( ! StrCompare( s + StrLen( s ) - 6, yearComp ) )
                 s[ StrLen( s ) - 6 ] = 0;
         }
     }

     StrCat( s, ", " );

     GetTimeFormat( &tf );

     TimeToAscii( dt.hour, dt.minute, tf, s + StrLen( s ) );

     if ( forceIncomplete || IsIncomplete( lastPart ) )
         StrCat( s, incompleteMarker );

     if ( *comment == '1' ) {
         StrNCat( s, autoMarker, MAX_FRIENDLY + 1 );
     }

     if ( comment[ 0 ] && comment[ 1 ] ) {
         StrCat( s, ": " );
         StrNCat( s, comment + 1, MAX_FRIENDLY + 1 );
     }

     return Duplicate( s );
}



static void DoList( void )
{
    Char** dirs;

    ClearListData();

    curCardVol = FindVolume( CARD, NULL );

    dirs = ScanDir( curCardVol, backupDir, &setCount, DIRS );

    if ( setCount < 0 ) {
        FatalError( altErrorScanning );
    }

    if ( 0 == setCount ) {
        sets = NULL;
    }
    else {
        UInt32 i;
        
        SysInsertionSort( dirs, setCount, sizeof( Char* ), CompareStrings, 0 );

        sets = SafeMemPtrNew( sizeof( *sets ) * setCount );
        friendlyNames = SafeMemPtrNew( sizeof( *friendlyNames ) * setCount );

        for ( i = 0 ; i < setCount ; i++ ) {
             sets[ i ].path = SafeMemPtrNew( StrLen( backupDir ) + 1 + StrLen( dirs[ i ] ) + 1 );

             StrPrintF( sets[ i ].path, "%s/%s", backupDir, dirs[ i ] );

             GetComment( sets[ i ].path );
             sets[ i ].automatic = ( comment[ 0 ] == '1' );;
             sets[ i ].friendly = GetFriendly( dirs[ i ], false );

             friendlyNames[ i ] = sets[ i ].friendly;

             SafeMemPtrFree( dirs[ i ] );
        }

        SafeMemPtrFree( dirs );
    }
}


static void DrawList( void )
{
    ListType* list = GetObjectPtr( lstSets );

    LstSetListChoices( list, friendlyNames, setCount );
    LstSetSelection( list, -1 /* ( Int16 )setCount -1 */ );
    if ( 0 < setCount )
        LstMakeItemVisible( list, setCount - 1 );
    LstDrawList( list );
}



#if 0
static void SelectBackup( void )
{
    select = selectBackup;

    selectSet = LstGetSelection( GetObjectPtr( lstSets ) );


    selectSourceDir = "/PALM_DM";
    selectSet       = -1;

    FrmPopupForm( frmSelect );
}
#endif


static void SelectRestore( void )
{
    select = selectRestore;

    selectSet = LstGetSelection( GetObjectPtr( lstSets ) );

    if ( selectSet < 0 ) {
        FrmAlert( altSelectSet );
        return;
    }

    selectSourceDir = sets[ selectSet ].path;

    FrmPopupForm( frmSelect );
}


static Boolean HandleMenu( UInt16 id )
{
    Boolean handled;

    handled = false;

    switch( id ) {
        case mnuSelectRange:
        case mnuUnselectRange:
            SelectRange( id == mnuSelectRange );
            handled = true;
            break;
        case mnuAbout:
            FrmPopupForm( frmAbout );
            handled = true;
            break;
        case mnuLicense:
            FrmHelp( strLicense );
            handled = true;
            break; 

#ifdef FLUSH
        case mnuFlush:
            DBCacheFlush();
            Init_frmMain();
            handled = true;
            break;
#endif
        case mnuImport: {
            FrmGotoForm( frmServer );
            handled = true;
            break;
        }
        case mnuExport: {
            Int16 sel;

            sel = GetSelection( lstSets );

            if ( 0 <= sel ) {
                if ( ! CheckBattery() )
                    FrmAlert( altBattery );
                else {
                    Export( curCardVol, sets[ sel ].path );
                    Init_frmMain();
                }
            }
            else
                FrmAlert( altSelectSet );

            handled = true;
            break;
        }

        case mnuEditComment: {
            Int16 sel;

            sel = GetSelection( lstSets );

            if ( 0 <= sel ) {
                GetComment( sets[ sel ].path );
                afterComment         = SAVE_COMMENT;
                afterCommentArgument = sel;
                FrmPopupForm( frmComment );
            }
            else
                FrmAlert( altSelectSet );

            handled = true;
            break;
        }

        case mnuTest: {
            LocalID id;
            DmOpenRef ref;

            id = DmFindDatabase( 0, "AletheSans 16px std" );
            ref = DmOpenDatabase( 0, id, dmModeReadWrite );
            DmRemoveResource( ref, DmNumResources( ref ) - 2 );
            DmCloseDatabase( ref );
            handled = true;
            break;
        }

        case mnuPrefs:
            FrmPopupForm( frmPrefs );
            handled = true;
            break;

        case mnuServerSettings:
            FrmPopupForm( frmFTP );
            handled = true;
            break;

        case mnuSelectRestore:
            SelectRestore();
            handled = true;
            break;

        case mnuLowMemRestore: {
            Int16 sel;
            
            sel = LstGetSelection( GetObjectPtr( lstSets ) );

            if ( 0 <= sel && Confirm( altLowMemRestore ) )
                Restore( sel, true );
            if ( sel < 0 )
                FrmAlert( altSelectSet );
            handled = true;
            break;
        }

        default:
            break;
     }
     
     return handled;
}



static void Init_frmMain( void )
{
    SetValue( chkInternalDrive, pref.internalDrive );
    Show( chkInternalDrive, 0 != FindVolume( NON_REMOVABLE, NULL ) );
    FrmDrawForm( FrmGetActiveForm() );
    DoList();
    DrawList();
}



static void AppendNumber( Char* s, UInt16 length, UInt16 n )
{
    s = s + StrLen( s ) + length;

    *s-- = 0;
    while ( length-- ) {
        *s-- = '0' + ( n % 10 );
        n /= 10;
    }
}


static void GetDateString( Char* s )
{
    DateTimeType dt;

    TimSecondsToDateTime( TimGetSeconds(), &dt );

    *s = 0;

    AppendNumber( s, 4, dt.year );
    AppendNumber( s, 2, dt.month );
    AppendNumber( s, 2, dt.day );
    AppendNumber( s, 2, dt.hour );
    AppendNumber( s, 2, dt.minute );
    AppendNumber( s, 2, dt.second );
}



Char* LastPart( Char* path )
{
    Char* p;

    p = path + StrLen( path ) - 1;

    while ( path <= p ) {
        if ( *p == '/' || *p == '\\' ) {
            return p+1;
        }
        p--;
    }

    return path;
}



Boolean HandleTimeServer( UInt32* newTimeP, Int32* deltaP )
{
    Boolean success;

    NetOpen();

    if ( pref.timeServer[ 0 ] )
        success = TimeServer( pref.timeServer, newTimeP, deltaP );
    else
        success = TimeServer( "time.nist.gov", newTimeP, deltaP );
        
    NetClose();

    return success;
}



static void Backup( Int16 sel )
{
    Char dirName[ MAX_FILENAME ];
    Char dirName0[ MAX_FILENAME ];
    Char correctName[ MAX_FILENAME ];
    Char* ds;
    Boolean success;
    LocalID id;

    if ( ! CheckBattery() ) {
        success = false;

        if ( ! scheduled )
            FrmAlert( altBattery );
        else
            ErrorInBackup();
        return;
    }

    StrCopy( dirName, backupDir );
    StrCat( dirName, "/" );

    ds = dirName + StrLen( dirName );
    GetDateString( ds );

    if ( 0 <= sel ) {
        if ( errNone != VFSFileRename( curCardVol, sets[ sel ].path, ds ) ) {
            if ( ! scheduled )
                FrmAlert( altErrorBackingUp );
            return;
        }
    }
    else {
        if ( ! MakeDir( curCardVol, dirName ) ) {
            if ( ! scheduled )
                FrmAlert( altErrorBackingUp );
            return;
        }
    }

    StrCopy( dirName0, dirName );
    StrCopy( correctName, ds );
    StrCat( ds, incompleteMarker );

    if ( errNone != VFSFileRename( curCardVol, dirName0, ds ) ) {
        if ( ! scheduled )
            FrmAlert( altErrorBackingUp );
        return;
    }

    DmCreateDatabase( 0, badBackupFileName, appCreator, badBackupFileType, false );
    DmSync();
    success = DoBackup( dirName, CARD, "/PALM_DM", HIDDEN, sel < 0, false,
                  scheduled ? "1" : comment, correctName,
                  pref.compression, NULL, 0, NULL );
    id = DmFindDatabase( 0, badBackupFileName );
    if ( id != NULL )
        DmDeleteDatabase( 0, id );

    if ( scheduled && success ) {
        if ( pref.exportAfterBackup && CheckBattery() )
            Export( curCardVol, dirName0 );

        pref.lastScheduledTime = TimGetSeconds();
        SavePrefs();

        if ( pref.setTime )
            HandleTimeServer( NULL, NULL );

        if ( pref.resetAfterScheduled )
            SysReset();
        else
            EvtEnqueueKey( vchrPowerOff, 0, commandKeyMask );
    }
    else if ( scheduled ) {
        ErrorInBackup();

        pref.lastScheduledTime = TimGetSeconds();
        SavePrefs();

        if ( pref.setTime )
            HandleTimeServer( NULL, NULL );
    }
    else if ( ! success ) {
        FrmAlert( altHadError ); 
    }

    if ( ! scheduled ) {
        Init_frmMain();
    }
}



static void Restore( Int16 sel, Boolean lowMemory )
{
    Char* lastPart;
    Boolean partial;

    lastPart = LastPart( sets[ sel ].path );

    partial = false;

    if ( IsIncomplete( lastPart ) ) {
        if ( Confirm( lowMemory ? altLowMemIncomplete : altIncomplete ) ) {
            partial = ! lowMemory;
        }
        else {
            return;
        }
    }

    DoBackup( "/PALM_DM", HIDDEN, sets[ sel ].path, CARD, lowMemory, partial, "", NULL, -1, NULL, 0, NULL );

    Init_frmMain();
}



static void SelectGoBackup( void )
{
//    DoBackup( "/PALM_DM", HIDDEN, sets[ selectSet ].path, CARD, false /* true */, partial, "", NULL, -1, NULL, 0, NULL );
}



static void SelectGoRestore( void )
{
    DoBackup( "/PALM_DM", HIDDEN, sets[ selectSet ].path, CARD, false /* true */, true, "", NULL, -1,
        fileList, fileCount, fileSelection );
}



Boolean HaveThisDirLocally( Char* dir )
{
     Int32 i;

     for ( i = 0 ; i < setCount ; i++ ) {
          Char* last;
          UInt16 len;
          
          last = LastPart( sets[ i ].path );
          len  = StrLen( last );

          if ( ! StrCompare( last, dir ) ||
               ( IsIncomplete( last ) &&
                 ! StrNCompare( last, dir, len - StrLen( incompleteMarker ) ) &&
                 dir[ len - StrLen( incompleteMarker ) ] == 0 ) )
              return true;
     }

     return false;
}


static void Delete( Int16 sel )
{
     Int32  count;
     UInt32 i;
     Char** names;

     names = ScanDir( curCardVol, sets[ sel ].path, &count, ALL );

     if ( count < 0 ) {
         FrmAlert( altErrorDeleting );
         return;
     }

     if ( names != NULL ) {
         DeleteUniqueFiles( curCardVol, sets[ sel ].path, names, count, NULL, 0 );
     }

     if ( errNone != VFSFileDelete( curCardVol, sets[ sel ].path ) ) {
         FrmAlert( altErrorDeleting );
     }

     DoList();
     FrmDrawForm( FrmGetActiveForm() );
     DrawList();

     for ( i = 0 ; i < count ; i++ )
          SafeMemPtrFree( names[ i ] );
     SafeMemPtrFree( names );
}



static Boolean Handler_frmMain( EventType* event )
{
    Boolean handled;
    static Boolean mainOpened = false;

    handled = false;

    switch (event->eType) {
        case winEnterEvent:
            if ( ! mainOpened || FrmGetActiveFormID() != frmMain )
                break;
            if ( redoList ) {
                redoList = false;
                DoList();
                DrawList();
                handled = true;
            }
            if ( redrawList ) {
                LstDrawList( GetObjectPtr( lstSets ) );
                redrawList = false;
                handled = true;
            }
            if ( nextAction != NOOP ) {
                EventType e;

                MemSet( &e, sizeof( e ), 0 );
                e.eType = DO_ACTION;
                EvtAddEventToQueue( &e );
            }
            break;

        case DO_ACTION:
            DoAction();
            handled = true;
            break;

        case frmCloseEvent:
            mainOpened = false;
            handled = false;
            break;

        case frmOpenEvent:
            if ( redoList ) {
                redoList = false;
                DoList();
                DrawList();
                handled = true;
            }
            Init_frmMain();
            mainOpened = true;
            handled = true;
            break;

        case ctlSelectEvent: {
            UInt16       id;
            Int16        sel;

            sel = GetSelection( lstSets );

            id = event->data.ctlEnter.controlID;

            switch( event->data.ctlEnter.controlID ) {
                case chkInternalDrive: {
                    Boolean oldValue;

                    oldValue = pref.internalDrive;

                    pref.internalDrive = GetValue( chkInternalDrive );

                    if ( 0 == FindVolume( CARD, NULL ) ) {
                        pref.internalDrive = oldValue;
                        SetValue( chkInternalDrive, oldValue );
                    }
                    else {
                        SavePrefs();

                        if ( ! pref.didWarnInternalDrive ) {
                            FrmAlert( altWarnInternal );
                            pref.didWarnInternalDrive = true;
                            SavePrefs();
                        }

                        MakeDirs( FindVolume( CARD, NULL ) );
                        DoList();
                        FrmDrawForm( FrmGetActiveForm() );
                        DrawList();
                    }
                    handled = true;
                    break;
                }
                case btnHelp:
                    FrmHelp( strFrmMainHelp );
                    handled = true;
                    break;
                case btnNewBackup:
                    afterComment = BACKUP;
                    afterCommentArgument = -1;
                    StrCopy( comment, "0manual backup" );
//                    Backup( -1 );
                    FrmPopupForm( frmComment );
                    handled = true;
                    break;
                case btnUpdate:
                    if ( 0 <= sel ) {
                        afterComment = BACKUP;
                        afterCommentArgument = sel;
                        GetComment( sets[ sel ].path );
//                        Backup( sel );
                        FrmPopupForm( frmComment );
                    }
                    else
                        FrmAlert( altSelectSet );

                    handled = true;
                    break;
                case btnRestore:
                    if ( 0 <= sel && Confirm( altRestore ) )
                        Restore( sel, false );
                    if ( sel < 0 )
                        FrmAlert( altSelectSet );
                    handled = true;
                    break;
                case btnDelete:
                    if ( 0 <= sel && Confirm( altDelete ) )
                        Delete( sel );
                    if ( sel < 0 )
                        FrmAlert( altSelectSet );
                    handled = true;
                    break;
            }
            break;
        }

        default:
            break;
    }
    return handled;
}



static Boolean Handler_frmAbout( EventType* event )
{
    Boolean handled;

    handled = false;

    switch (event->eType) {
        case frmOpenEvent:
            FrmDrawForm( FrmGetActiveForm() );
            handled = true;
            break;

        case ctlSelectEvent: {
            switch( event->data.ctlEnter.controlID ) {
                case btnOK:
                    FrmReturnToForm( 0 );
                    handled = true;
                    break;
                case btnLicense:
                    FrmHelp( strLicense );
                    handled = true;
                    break;
            }
            break;
        }

        default:
            break;
    }
    return handled;
}



void Show( UInt16 id, Boolean state )
{
    UInt16 index = GetObjectIndex( id );
    FormType* currentForm = FrmGetActiveForm();

    if ( state )
        FrmShowObject( currentForm, index );
    else
        FrmHideObject( currentForm, index );
    if ( FrmGetObjectType( currentForm, index ) == frmControlObj )
        CtlSetUsable( GetObjectPtr( id ), state);
}


static void UpdateAdvancedOptionDisplay( void )
{
    Show( chkResetAfterScheduled, pref.daily );
    Show( chkSetTime, pref.daily );
}



static void UpdateOptionDisplay( void )
{
    static Char timeLabel[ 40 ];
    TimeFormatType tf;

    Show( trgCycle, pref.daily );
    Show( lblCycle, pref.daily );
    Show( trgTime, pref.daily );
    Show( lblAt,   pref.daily );
    Show( lblWithKey, pref.doCrypt );
    Show( trgKey,     pref.doCrypt );
    
    if ( pref.doCrypt ) {
        CtlSetLabel( GetObjectPtr( trgKey ),
             pref.haveKey ? "assigned" : "prompt" );
    }

#if 0
    if ( pref.daily && PwdExists() ) {
        Char password[ MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE ];
        PasswordDecrypt( password, pref.devicePassword );

        if ( ! PwdVerify( password ) ) {
            pref.haveDevicePassword = false;
            MemZero( pref.devicePassword, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE );
            SavePrefs();
        }
        MemZero( password, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE );

        Show( lblPassword, true );
        Show( trgPassword, true );
        CtlSetLabel( GetObjectPtr( trgPassword ),
             pref.haveDevicePassword ? "assigned" : "unassigned" );
    }
    else {
        if ( pref.haveDevicePassword ) {
            MemSet( pref.devicePassword, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE, 0 );
            pref.haveDevicePassword = false;
            SavePrefs();
        }
        Show( lblPassword, false );
        Show( trgPassword, false );
    }
#endif    

    GetTimeFormat( &tf );

    TimeToAscii( pref.scheduleHour, pref.scheduleMinute, tf, timeLabel );

    CtlSetLabel( GetObjectPtr( trgTime ), timeLabel );
}



#if 0
static Boolean PasswordVerify( Int16 button, Char* attempt )
{
    if ( button != 0 )
        return true;

    return PwdVerify( attempt );
}
#endif



Boolean KeyVerify( Int16 button, Char* attempt )
{
    return button != 0 || *attempt;
}



static Boolean Handler_frmPrefs( EventType* event )
{
    Boolean handled;

    handled = false;

    switch (event->eType) {
        case frmOpenEvent: {
            UInt16 i;

            UpdateOptionDisplay();
            FrmDrawForm( FrmGetActiveForm() );
            SetList( trgCycle, lstCycle, pref.cycle - 1 );
            SetList( trgCompression, lstCompression, pref.compression + 1 );
#ifdef FLUSH
            SetList( trgFlush, lstFlush, pref.flush );
#endif
            SetValue( chkDaily, pref.daily );
            SetValue( chkEncrypt, pref.doCrypt );

            for ( i = 0 ; i < NUM_BOOLEAN_OPTIONS ; i++ )  {
                 if ( ! booleanOptions[ i ].advanced )
                     SetValue( booleanOptions[ i ].id, *PREF_BOOLEAN_PTR( booleanOptions[ i ].offset ) );
            }

            handled = true;
            break;
        }

        case ctlSelectEvent: {
            switch( event->data.ctlEnter.controlID ) {
                case btnAdvanced:
                    FrmPopupForm( frmAdvanced );
                    handled = true;
                    break;
                case btnOK:
                    if ( pref.daily && pref.doCrypt && ! pref.haveKey ) {
                        FrmAlert( altNeedKey );
                    }
                    else {
                        FrmReturnToForm( 0 );
                    }    
                    handled = true;
                    break;
#if 0
                case trgPassword: {
                    Char password[ MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE ];
                    password[ 0 ] = 0;
                    switch ( FrmCustomResponseAlert( altPassword, " your device unlock", "", "",
                                 password, MAX_PASSWORD_SIZE, PasswordVerify ) ) {
                        case 0:
                            MemSet( pref.devicePassword, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE, 0 );
                            PasswordEncrypt( pref.devicePassword, password );
                            MemZero( password, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE );
                            pref.haveDevicePassword = true;
                            break;
                        case 1:
                            pref.haveDevicePassword = 0;
                            MemSet( pref.devicePassword, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE, 0 );
                            break;
                    }
                    SavePrefs();
                    UpdateOptionDisplay();
                    handled = true;
                    break;
                }
#endif                
                case trgKey: {
                    Char password[ MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE ];
                    password[ 0 ] = 0;
                    switch ( FrmCustomResponseAlert( altPassword, " encryption", "", "",
                                 password, MAX_PASSWORD_SIZE, KeyVerify ) ) {
                        case 0:
                            MemSet( pref.key, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE, 0 );
                            PasswordEncrypt( pref.key, password );
                            MemSet( password, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE, 0 );
                            pref.haveKey = true;
                            break;
                        case 1:
                            pref.haveKey = 0;
                            MemSet( pref.key, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE, 0 );
                            break;
                    }
                    SavePrefs();
                    UpdateOptionDisplay();
                    handled = true;
                    break;
                }
                case chkDaily:
                    pref.daily = GetValue( chkDaily );
                    pref.forceNextAlarm = 0;
                    SavePrefs();
                    UpdateSchedule();
                    UpdateOptionDisplay();
                    handled = true;
                    break;
                case chkEncrypt:
                    pref.doCrypt = GetValue( chkEncrypt );
                    if ( pref.doCrypt && pref.compression < 0 ) {
                        pref.compression = 0;
                        SetList( trgCompression, lstCompression, pref.compression + 1 );
                        CtlDrawControl( GetObjectPtr( trgCompression ) );
                    }
                    if ( ! pref.doCrypt ) {
                        pref.haveKey = 0;
                        MemSet( pref.key, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE, 0 );
                    }  
                    SavePrefs();
                    UpdateOptionDisplay();
                    handled = true;
                    break;
#if 0
                case chkSlowUpdate:
                    pref.slowUpdate = GetValue( chkSlowUpdate );
                    SavePrefs();
                    handled = true;
                    break;
#endif
                case trgTime: {
                    UInt16 h;
                    UInt16 m;

                    h = pref.scheduleHour;
                    m = pref.scheduleMinute;

                    pref.forceNextAlarm = 0;
                    SavePrefs();

                    if ( SelectOneTime( &h, &m, "Scheduled backup time" ) ) {
                        pref.scheduleHour = h;
                        pref.scheduleMinute = m;
                        SavePrefs();
                        UpdateSchedule();
                        UpdateOptionDisplay();
                    }
                }

                default: {
                    UInt16 i;
                    
                    for ( i = 0 ; i < NUM_BOOLEAN_OPTIONS ; i++ ) {
                        if ( event->data.ctlEnter.controlID == booleanOptions[ i ].id ) {
                            *PREF_BOOLEAN_PTR( booleanOptions[ i ].offset ) = 
                                GetValue( event->data.ctlEnter.controlID );
                            SavePrefs();
                            handled = true;
                            break;
                        }
                    }
                    break;
                }
            }
            break;
        }

        case popSelectEvent:
            if ( event->data.popSelect.selection != noListSelection ) {
                switch ( event->data.popSelect.controlID ) {
                    case trgCompression: {
                        pref.compression = ( Int8 )GetSelection( lstCompression ) - 1;
                        if ( pref.compression < 0 && pref.doCrypt ) {
                            pref.compression = 0;
                            SetList( trgCompression, lstCompression, pref.compression + 1 );
                            CtlDrawControl( GetObjectPtr( trgCompression ) );
                        }
                        SavePrefs();
                        break;
                    }
#ifdef FLUSH
                    case trgFlush: {
                        pref.flush = GetSelection( lstFlush );
                        SavePrefs();
                        break;
                    }
#endif
                    case trgCycle: {
                        pref.cycle = 1 + GetSelection( lstCycle );
                        SavePrefs();
                        UpdateSchedule();
                        break;
                    }
                }
            }
            break;

        default:
            break;
    }
    return handled;
}



static Boolean Handler_frmAdvanced( EventType* event )
{
    Boolean handled;

    handled = false;

    switch (event->eType) {
        case frmOpenEvent: {
            UInt16 i;
            UInt16 n;
            ListType* timeServers;

            timeServers = GetObjectPtr( lstTimeServer );

            n           = LstGetNumberOfItems( timeServers );

            for ( i = 0 ; i < n ; i++ ) {
                 if ( ! StrCaselessCompare( pref.timeServer, LstGetSelectionText( timeServers, i ) ) ) {
                     break;
                 }
            }

            if ( i == n )
                i = 0;

            SetList( trgTimeServer, lstTimeServer, i );

            StrCopy( pref.timeServer, LstGetSelectionText( timeServers, i ) );
            SavePrefs();

            for ( i = 0 ; i < NUM_BOOLEAN_OPTIONS ; i++ )
                 if ( booleanOptions[ i ].advanced )
                     SetValue( booleanOptions[ i ].id, *PREF_BOOLEAN_PTR( booleanOptions[ i ].offset ) );

            UpdateAdvancedOptionDisplay();

            FrmDrawForm( FrmGetActiveForm() );

            handled = true;
            break;
        }

        case popSelectEvent:
            if ( event->data.popSelect.selection != noListSelection ) {
                switch ( event->data.popSelect.controlID ) {
                    case trgTimeServer:  {
                        ListType* timeServers;

                        timeServers = GetObjectPtr( lstTimeServer );

                        StrCopy( pref.timeServer, LstGetSelectionText( timeServers, event->data.popSelect.selection ) );
                        SavePrefs();

                        break;
                    }
                }
            }
            break;



        case ctlSelectEvent: {
            switch( event->data.ctlEnter.controlID ) {
                case btnSetTime: {
                    Boolean success;
                    Int32   delta;
                    UInt32  newTime;

                    Show( lblConnecting, true );

                    success = HandleTimeServer( &newTime, &delta);

                    Show( lblConnecting, false );

                    if ( success ) {
                        static Char msg[ 80 ];
                        static Char ampm[ 3 ];
                        DateFormatType df;
                        TimeFormatType tf;
                        DateTimeType   t;
                        Char           separator;
                        UInt16         len;

                        TimSecondsToDateTime( newTime, &t );

                        GetTimeFormat( &tf );

                        TimeToAscii( t.hour, t.minute, tf, msg );

                        separator = ':';
                        if ( NULL != StrChr( msg, ',' ) )
                            separator = ',';
                        else if ( NULL != StrChr( msg, '.' ) )
                            separator = '.';

                        len = StrLen( msg );

                        *ampm = 0;

                        if ( 2 < len && ( ! StrCaselessCompare( msg + len - 2, "am" ) ||
                                 ! StrCaselessCompare( msg + len - 2, "pm" ) ) ) {
                                StrCopy( ampm, msg + len - 2 );
                                if ( 3 < len && msg[ len - 3 ] == ' ' )
                                    len -= 3;
                                else
                                    len -= 2;
                                msg[ len ] = 0;
                        }

                        StrPrintF( msg + len, "%c%d%d%s, ", separator, t.second/10, t.second%10, ampm );
                        len = StrLen( msg );

                        GetDateFormat( &df );

                        DateToAscii( t.month, t.day, t.year, df, msg + len );

                        len = StrLen( msg );

                        StrPrintF( msg + len, " (%s%ld sec) ", ( 0 <= delta ) ? "+" : "", delta );

                        FrmCustomAlert( altTimeSuccess, msg, "", "" );
                    }
                    else
                        FrmAlert( altTimeError );

                    handled = true;
                    break;
                }

                case btnOK:
                    FrmReturnToForm( 0 );
                    handled = true;
                    break;

                default: {
                    UInt16 i;

                    for ( i = 0 ; i < NUM_BOOLEAN_OPTIONS ; i++ ) {
                        if ( event->data.ctlEnter.controlID == booleanOptions[ i ].id ) {
                            *PREF_BOOLEAN_PTR( booleanOptions[ i ].offset ) =
                                GetValue( event->data.ctlEnter.controlID );
                            SavePrefs();
                            UpdateAdvancedOptionDisplay();
                            handled = true;
                            break;
                        }
                    }

                    break;
                }
            }
            break;
        }

        default:
            break;
    }
    return handled;
}



void CopyFromField( Char* dest, UInt16 id, UInt16 bufLen )
{
    Char* s;
    
    if ( bufLen == 0 )
        return;

    MemSet( dest, bufLen, 0 );

    s = FldGetTextPtr( GetObjectPtr( id ) );

    if ( s == NULL )
        return;

    StrNCopy( dest, s, bufLen - 1 );
}



void SetField( UInt16 id, Char* s )
{
    FieldType* f;

    f = GetObjectPtr( id );

    FldDelete( f, 0, 0xFFFF );
    FldInsert( f, s, StrLen( s ) );
}



static void DoAction( void )
{
    ActionType old;

    old          = nextAction;
    nextAction   = NOOP;

    switch ( old ) {
        case BACKUP: {
            Backup( nextActionArgument );
            break;
        }
        case NOOP:
            break;
        case SAVE_COMMENT:
            SaveComment( curCardVol, sets[ nextActionArgument ].path, comment );
            DoList();
            DrawList();
            break;
    }
}



static Coord FrmSelectStringWidth( Int16 item )
{
    Char   fixed[ MAX_FILENAME ];

    FixName( fixed, LstGetSelectionText( GetObjectPtr( lstFiles ), item ) );

    return FntCharsWidth( fixed, StrLen( fixed ) );
}



static void FrmSelectScrollTo( Char c )
{
    Int16 i;
    UInt16 n;
    ListType* list;
    
    list = GetObjectPtr( lstFiles );

    n = LstGetNumberOfItems( list );

    for ( i = 0 ; i < n ; i++ ) {
         Char   fixed[ MAX_FILENAME ];

         FixName( fixed, LstGetSelectionText( list, i ) );

         if ( 0 == StrNCaselessCompare( &c, fixed, 1 ) )
             break;

         if ( StrNCaselessCompare( &c, fixed, 1 ) < 0 ) {
             i--;
             break;
         }
    }

    if ( i == n )
        i = n - 1;
    if ( i < 0 )
        i = 0;
    LstMakeItemVisible( list, i );
    LstDrawList( list );
}



static Boolean RangeVerify( Int16 button, Char* attempt )
{
    if ( button != 0 )
        return true;
    if ( ! attempt[ 0 ] )
        return false;
    if ( attempt[ 0 ] == '-' )
        return attempt[ 1 ] && ! attempt[ 2 ];
    if ( ! attempt[ 1 ] )
        return true;
    if ( attempt[ 1 ] != '-' )
        return false;
    return true;
}


Char LowerChar( Char c )
{
    Char string[ 2 ];
    Char outString[ 2 ];

    string[ 0 ] = c;
    string[ 1 ] = 0;

    StrToLower( outString, string );
    
    return outString[ 0 ];
}


void SelectChar( Char c, Boolean value )
{
    UInt16 i;
    
    for ( i = 0 ; i < fileCount ; i++ ) {
         Char c1;
         
         c1 = LowerChar( fileList[ i ][ 0 ] );

         if ( c == '.' ) {
             if ( c1 < 'a' || 'z' < c1 )
                 fileSelection[ i ] = value;
         }
         else if ( c1 == c )
                 fileSelection[ i ] = value;
    }
}


void SelectBetween( Char a, Char b, Boolean value )
{
    a = LowerChar( a );
    b = LowerChar( b );

    if ( b < a )
        return;

    if ( a == b ) {
        SelectChar( a, value );
    }
    else {
        Char i;

        if ( b == '.' ) {
            SelectChar( '.', value );
            b = 'z';
        }
        if ( a == '.' ) {
            SelectChar( '.', value );
            a = 'a';
        }

        if ( b < a )
            return;

        for ( i = a ; i <= b ; i++ )
            SelectChar( i, value );
    }
}


static void SelectRange( Boolean select )
{
    Char range[ 4 ];
    UInt16 i;

    if ( 0 != FrmCustomResponseAlert( altSelectRange, select ? "Select" : "Unselect", select ? "select" : "unselect", "",
                   range, 4, RangeVerify ) ) {
        return;
    }

    if ( range[ 0 ] == '-' ) {
        SelectBetween( '.', range[ 1 ], select );
    }
    else if ( range[ 1 ] == '-' ) {
        SelectBetween( range[ 0 ], range[ 2 ] ? range[ 2 ] : 'z', select );
    }
    else {
        SelectBetween( range[ 0 ], range[ 0 ], select );
    }
    
    LstDrawList( GetObjectPtr( lstFiles ) );
}


static Boolean Handler_frmSelect( EventType* event )
{
    Boolean handled;

    handled = false;

    switch (event->eType) {
        case frmOpenEvent:
            FrmDrawForm( FrmGetActiveForm() );

            CtlSetLabel( GetObjectPtr( btnSelectGo ), selectGo[ select ].label );
            Show( btnSelectGo, true );

            SetupFileList();

            handled = true;
            break;

        case ctlSelectEvent: {
            switch( event->data.ctlEnter.controlID ) {
                case btnSelectGo:
                    selectGo[ select ].Action();
                    FrmReturnToForm( 0 );
                    handled = true;
                    break;
                case btnCancel:
                    FrmReturnToForm( 0 );
                    handled = true;
                    break;
                case btnAll:
                    MemSet( fileSelection, sizeof( Boolean ) * fileCount, true );
                    LstDrawList( GetObjectPtr( lstFiles ) );
                    handled = true;
                    break;
                case btnNone:
                    MemSet( fileSelection, sizeof( Boolean ) * fileCount, false );
                    LstDrawList( GetObjectPtr( lstFiles ) );
                    handled = true;
                    break;
            }
            break;
        }

        case keyDownEvent: {
            if ( event->data.keyDown.modifiers & commandKeyMask ) {
                switch ( event->data.keyDown.chr ) {
                    case vchrRockerUp:
                    case chrPageUp:
                        LstScrollList( GetObjectPtr( lstFiles ), winUp,
                            LstGetVisibleItems( GetObjectPtr( lstFiles ) ) - 1 );
                        handled = true;
                        break;
                    case vchrRockerDown:
                    case chrPageDown:
                        LstScrollList( GetObjectPtr( lstFiles ), winDown,
                            LstGetVisibleItems( GetObjectPtr( lstFiles ) ) - 1 );
                        handled = true;
                        break;
                    default:
                        break;
                }
            }
            else {
                if ( 0 == (0xFF00 & event->data.keyDown.chr ) ) {
                    FrmSelectScrollTo( event->data.keyDown.chr );
                    handled = true;
                }
            }
            break;
        }

        case lstSelectEvent: {
            Int16 item;
            RectangleType bounds;

            item = event->data.lstSelect.selection;

            FrmGetObjectBounds( FrmGetActiveForm(),
                GetObjectIndex( event->data.lstSelect.listID ), &bounds );

            switch ( event->data.lstSelect.listID ) {
                case lstFiles:
                    if ( 0 <= item &&
                         event->screenX < bounds.topLeft.x + CHECKMARK_SIZE +
                             FrmSelectStringWidth( item ) ) {
                        fileSelection[ item ] = ! fileSelection[ item ];
                    }
                    LstSetSelection( GetObjectPtr( lstFiles ), -1 );
                    handled = true;
                    break;
            }
            break;
        }

        default:
            break;
    }
    return handled;
}



static void EventLoop( void )
{
    EventType event;
    Boolean   handled;

    terminate = false;

    do {
       	EvtGetEvent( &event, evtWaitForever );

       	handled = SysHandleEvent( &event );

       	if ( ! handled ) {
       	    Err err;

       	    handled = MenuHandleEvent( NULL, &event, &err );
       	}

       	if ( ! handled ) {
            switch ( event.eType ) {
                case frmLoadEvent: {
                    UInt16    formID;
                    FormType* form;

                    formID = event.data.frmLoad.formID;

                    form = FrmInitForm( formID );
                    FrmSetActiveForm( form );

                    switch ( formID ) {
                        case frmMain:
                            SetResizePolicy( frmMain );
                            FrmSetEventHandler( form, Handler_frmMain );
                            break;
                        case frmServer:
                            SetResizePolicy( frmServer );
                            FrmSetEventHandler( form, Handler_frmServer );
                            break;
                        case frmAbout:
                            SetResizePolicy( frmAbout );
                            FrmSetEventHandler( form, Handler_frmAbout );
                            break;
                        case frmPrefs:
                            SetResizePolicy( frmPrefs );
                            FrmSetEventHandler( form, Handler_frmPrefs );
                            break;
                        case frmAdvanced:
                            SetResizePolicy( frmPrefs );
                            FrmSetEventHandler( form, Handler_frmAdvanced );
                            break;
                        case frmComment:
                            SetResizePolicy( frmComment );
                            FrmSetEventHandler( form, Handler_frmComment );
                            break;
                        case frmFTP:
                            SetResizePolicy( frmFTP );
                            FrmSetEventHandler( form, Handler_frmFTP );
                            break;
                        case frmSelect:
                            SetResizePolicy( frmSelect );
                            FrmSetEventHandler( form, Handler_frmSelect );
                            break;
                    }
                    handled = true;
                }
                case menuEvent:
                    handled = HandleMenu( event.data.menu.itemID );
                    break;
                default:
                    handled = ResizeHandleEvent( &event );
                    if ( ! handled )
                        handled = FrmDispatchEvent( &event );
            }
       	}
    } while ( event.eType != appStopEvent && ! terminate );
    
    FrmCloseAllForms();
}



static Boolean ApplicationStart( void )
{
    LoadPrefs();

    if ( ! GetCard() ) {
        FrmAlert( altNoCard );
        return false;
    }

    InitZip();
    UpdateSchedule();

    InitializeResizeSupport( resizeIndex );
    LoadResizePrefs( ( UInt32 )appCreator, PREF_ID_RESIZE );

    FrmGotoForm(frmMain);

    return true;
}


static void ApplicationStop( void )
{
//    Message(":Cleaning up compression");
    CleanupZip();
//    Message(":Cleaning up backup data");
    CleanupBackup();
//    Message(":Cleaning up list data");
    ClearListData();
//    Message(":Cleaning up file list");
    ClearFileList();
//    Message(":Saving resize data");
    SaveResizePrefs( appCreator, PREF_ID_RESIZE, 0 );
//    Message(":Terminating resize support");
    TerminateResizeSupport();
//    Message(":Done");
}



static Boolean Warnings( void )
{
    PreferenceType* p;
    UInt16         opt;
    UInt32         errorValue;

    p = MemPtrNew( sizeof( PreferenceType ) );

    if ( p == NULL )
        return false;

    LoadPrefs0( p );

    if ( ! p->didLicense ) {
        for( ;; ) {
            opt = FrmAlert( altLicense );

            if ( opt == 0 ) {
                SafeMemPtrFree( p );
                return false;
            }
            else if ( opt == 2 ) {
                p->didLicense = 1;
                SavePrefs0( p );
                break;
            }
            else if ( opt == 1 ) {
                FrmHelp( strLicense );
            }
        }
    }

    if ( ! p->didCompatibility ) {
        FrmAlert( altNoCrossCompatibility );
        p->didCompatibility = true;
        SavePrefs0( p );
    }

    if ( ! p->didHelp ) {
        FrmHelp( strFrmMainHelp );
        p->didHelp = true;
        SavePrefs0( p );
    }

    if ( ! p->newCyclingWarned ) {
        p->newCyclingWarned = true;
        SavePrefs0( p );

        if ( p->daily ) {
            FrmAlert( altNewCycling );
        }
    }

    SafeMemPtrFree( p );

    if ( errNone == FtrGet( appCreator, FTR_FROM_ERROR, &errorValue ) ) {
        UInt16  card;
        LocalID id;

        FtrUnregister( appCreator, FTR_FROM_ERROR );

        if ( FindMe( &card, &id ) )
            AttnForgetIt( card, id, errorValue );

        FrmAlert( altHadError );
    }

    return true;
}



static Boolean HaveNVFS( void )
{
    Err err;
    UInt32 value;

    err = FtrGet( sysFtrCreator, sysFtrNumDmAutoBackup, &value );

    return ( err == errNone && value != 0 );
}



static void CrucialStart( void )
{
    InitZip();
    LoadPrefs();
    DoList();
}



static void CrucialStop( void )
{
    CleanupBackup();
    CleanupZip();
    ClearListData();
}


static void AttentionError( UInt16 msgID )
{
    UInt16 card;
    LocalID id;

    if ( ! FindMe( &card, &id ) )
        return;

    AttnGetAttention( card, id, msgID, NULL, kAttnLevelInsistent,
       kAttnFlagsUseUserSettings, 0, 1 );
}



static void ErrorInBackup( void )
{
    AttentionError( strBadBackup );
}


static void AutoBackup( void )
{
    Int32 i;
    Int32 countDown;
    WinHandle newWindow;
    WinHandle oldWindow;
    RectangleType bounds;
    Err       err;
    WinHandle save;
    UInt32    value;

    bounds.topLeft.x = 0;
    bounds.topLeft.y = 0;

    WinGetDisplayExtent( &bounds.extent.x, &bounds.extent.y );

    newWindow = WinCreateWindow( &bounds, noFrame, false, true, &err );

    if ( newWindow == NULL || err != errNone ) {
        ErrorInBackup();
        return;
    }

    oldWindow = WinGetDrawWindow();
    WinSetDrawWindow( newWindow );

    if ( scheduled && errNone == FtrGet( appCreator, FTR_UI_SWITCH, &value ) &&
         ! value ) {
        save = WinSaveBits( &bounds, &err );
        if ( save == NULL || err != errNone ) {
            if ( oldWindow != NULL )
                WinSetDrawWindow( oldWindow );
            WinDeleteWindow( newWindow, false );
            ErrorInBackup();
            return;
        }
    }
    else {
        save = NULL;
    }

    scheduled = true;

    LoadPrefs();
    if ( pref.dimOnScheduled )
        LCDState( false );
    CrucialStart();

    countDown = pref.cycle;

    for ( i = setCount - 1 ; 0 <= i ; i-- ) {
         if ( sets[ i ].automatic ) {
             countDown--;
             if ( countDown == 0 ) {
                 break;
             }
         }
    }

    Backup( i ); // If we don't have enough automatic backups, this will be -1.

    CrucialStop();
    if ( pref.dimOnScheduled )
        LCDState( true );

    scheduled = false;

    if ( save != NULL ) {
        WinRestoreBits( save, 0, 0 );
        MemPtrFree( save );
    }
    if ( oldWindow != NULL )
        WinSetDrawWindow( oldWindow );
    WinDeleteWindow( newWindow, false );
}



#if 0
static void FeedChar( UInt16 chr, UInt16 keyCode, UInt16 modifiers )
{
    UInt32    maxT = TimGetTicks() + SysTicksPerSecond() / 10;

    EvtEnqueueKey( chr, keyCode, modifiers );

    while ( maxT < TimGetTicks() ) {
        EventType e;
        Boolean handled;
        Boolean myEvent;

        EvtGetEvent( &e, SysTicksPerSecond() / 10 );

        myEvent = ( e.eType == keyDownEvent && e.data.keyDown.chr == chr &&
                    e.data.keyDown.keyCode == keyCode && e.data.keyDown.modifiers == modifiers );

        handled = SysHandleEvent( &e );

        if ( ! handled ) {
            Err err;

            handled = MenuHandleEvent( NULL, &e, &err );
        }

        if ( ! handled )
            handled = FrmDispatchEvent( &e );

        if ( myEvent )
            break;
    }
}



static void FeedString( Char* s )
{
    while ( *s ) {
        FeedChar( ( UInt8 )*s++, 0, 0 );
    }
}




static Boolean UnlockAsNeeded( void )
{
    PreferenceType* p;
    Char password[ MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE ];
    password[ 0 ] = 0;

    if ( ! PrefGetPreference( prefDeviceLocked ) )
        return true;

    p = MemPtrNew( sizeof( PreferenceType ) );

    if ( p == NULL )
        return false;

    LoadPrefs0( p );

    if ( ! p->haveDevicePassword ) {
        SafeMemPtrFree( p );
        return false;
    }

    PasswordDecrypt( password, p->devicePassword );

    if ( ! PwdVerify( password ) ) {
        SafeMemPtrFree( p );
        return false;
    }

    p->needLock = true;
    SavePrefs0( p );

    FeedString( password );
    FeedChar( '\n', 0, 0 );

    SafeMemPtrFree( p );

    return true;
}



void LockIfNeeded1( void )
{
      PreferenceType* p;

      p = MemPtrNew( sizeof( PreferenceType ) );
      if ( p != NULL ) {
          LoadPrefs0( p );
          LockIfNeeded( p );
          SafeMemPtrFree( p );
      }
 }
#endif


#if 0
void DoContinue( UInt16 almProcCmd, SysAlarmTriggeredParamType* paramP )
{
    EvtEnqueueKey( vchrConfirm, 0, commandKeyMask );
}
#endif


static Boolean HotsyncActive( void )
{
    UInt32 value;

    return errNone == FtrGet( appCreator, FTR_SYNC, &value );
}


static Boolean NVBackupBusy( void )
{
    UInt32 value;

    return errNone == FtrGet( appCreator, FTR_BUSY, &value );
}


static void ContinueScheduled( PreferenceType* p )
{
    UInt16  answer;

    LoadPrefs0( p );

    if ( HotsyncActive() || NVBackupBusy() ) {
        p->forceNextAlarm = TimGetSeconds() + SNOOZE_TIME;
        return;
    }

    p->forceNextAlarm = 0;
    SavePrefs0( p );

    answer = SchedulerDialog();
    if ( answer == 0 ) {
        LocalID id;
        UInt16  card;

        if ( FindMe( &card, &id ) ) {
            UInt32 result;

            if ( PrefGetPreference( prefDeviceLocked ) ) {
                FtrSet( appCreator, FTR_UI_SWITCH, 0 );
                SysAppLaunch( card, id, sysAppLaunchFlagNewGlobals, AUTO_BACKUP, NULL, &result );
            }
            else {
                FtrSet( appCreator, FTR_UI_SWITCH, 1 );
                SysUIAppSwitch( card, id, AUTO_BACKUP, NULL );
            }
        }
    }
    else if ( answer == 1 ) {
        p->forceNextAlarm = TimGetSeconds() + SNOOZE_TIME;
        SavePrefs0( p );
    }
}


static void DoRegister( void )
{
    UInt16 card;
    LocalID id;
    
    if ( ! FindMe( &card, &id ) )
        return;

    SysNotifyRegister( card, id, sysNotifySyncStartEvent, NULL, 0, 0 );
    SysNotifyRegister( card, id, sysNotifySyncFinishEvent, NULL, 0, 0 );
    SysNotifyRegister( card, id, sysNotifyTimeChangeEvent, NULL, 0, 0 );
}


void HandleNotification( SysNotifyParamType* np )
{
    switch ( np->notifyType ) {
        case sysNotifySyncStartEvent:
            FtrSet( appCreator, FTR_SYNC, 1 );
            break;
        case sysNotifySyncFinishEvent:
            FtrUnregister( appCreator, FTR_SYNC );
            break;
        case sysNotifyTimeChangeEvent: {
            PreferenceType* p;

            p = MemPtrNew( sizeof( PreferenceType ) );

            if ( p != NULL ) {
                LoadPrefs0( p );
                UpdateSchedule0( p, *( Int32* )( np->notifyDetailsP ) );
                MemPtrFree( p );
            }
            break;
        }
//        case appCreator:
//            FtrSet( appCreator, FTR_BUTTON, _ProcessButtons( np->notifyDetailsP ) );
//            break;

    }
}



void DrawAttentionItem( AttnLaunchCodeArgsType* p )
{
    MemHandle h;
    
    h = DmGetResource( 'tSTR', p->userData );
    
    if ( h != NULL ) {
        Char* s;
        
        s = MemHandleLock( h );
        
        WinDrawChars( s, StrLen( s ), p->commandArgsP->drawDetail.bounds.topLeft.x,
                                      p->commandArgsP->drawDetail.bounds.topLeft.y );

        MemHandleUnlock( h );
        DmReleaseResource( h );
    }
}



UInt32 PilotMain( UInt16 command, void* cmdPBP, UInt16 flags )
{
    switch ( command ) {
        case sysAppLaunchCmdNormalLaunch: {
            Err err;

            if ( ! HaveNVFS() ) {
                FrmAlert( altNoNVFS );
                break;
            }
            if ( ! Warnings() )
                break;
            ErrTry {
                err = errNone;
                if ( ApplicationStart() ) {
                    EventLoop();
                    ApplicationStop();
                }
            }
            ErrCatch( err ) {
                ApplicationStop();
                if ( err != errNone )
                    return err;
                break;
            } ErrEndCatch
            break;
        }
        case AUTO_BACKUP:  {
//            if ( ! HaveNVFS() )
//                break;

#if 0
            if ( ! HaveNVFS() ) {
                LockIfNeeded1();
                break;
            }
#endif

            AutoBackup();
//            LockIfNeeded1();
            break;
        }
        case sysAppLaunchCmdDisplayAlarm: {
            PreferenceType* p;

            p = MemPtrNew( sizeof( PreferenceType ) );

            if ( p == NULL )
                break;

            LoadPrefs0( p );

//            if ( PrefGetPreference( prefDeviceLocked ) ) {
//                SafeMemPtrFree( p )
//                break;
//            }

            if ( p != NULL ) {
                ContinueScheduled( p );
                UpdateSchedule0( p, 0 );
                SafeMemPtrFree( p );
            }
            break;
        }
        case sysAppLaunchCmdSystemReset: {
            UInt16 card;
            LocalID id;

            if ( FindBadBackupFile( &card, &id ) ) {
                DmDeleteDatabase( card, id );
                ErrorInBackup();
            }
        }
        // Fall through to next case!
        case sysAppLaunchCmdSyncNotify:
        {
            PreferenceType* p;

            p = MemPtrNew( sizeof( PreferenceType ) );

            if ( p != NULL ) {
                FtrSet( appCreator, FTR_NOT_WARM, 1 );
                DoRegister();
                LoadPrefs0( p );
                UpdateSchedule0( p, 0 );
                MemPtrFree( p );
            }
            break;
        }
        case sysAppLaunchCmdAttention: {
            AttnLaunchCodeArgsType* p = cmdPBP;

            switch ( p->command ) {
                case kAttnCommandDrawDetail:
                case kAttnCommandDrawList:
                    DrawAttentionItem( p );
                    break;
                case kAttnCommandGoThere: {
                    LocalID id;
                    UInt16  card;

                    if ( FindMe( &card, &id ) ) {
                        FtrSet( appCreator, FTR_FROM_ERROR, p->userData );
                        SysUIAppSwitch( card, id, sysAppLaunchCmdNormalLaunch, NULL );
                    }

                    break;
                }
            }
            break;
        }

        case sysAppLaunchCmdNotify: {
            Boolean handled = false;

            if ( flags & sysAppLaunchFlagSubCall )
                handled = HandleResizeNotification( ( (SysNotifyParamType *)cmdPBP )->notifyType );

            if ( ! handled )
                HandleNotification( (SysNotifyParamType *)cmdPBP );

            break;
        }
        default:
            break;
    }
    return 0;
}


