/* $Id: zip.h,v 1.7 2006/05/26 00:16:32 arpruss Exp $ */
/*****************************************************************************

    unzip
    Copyright (C) 2005 Alexander R Pruss

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*****************************************************************************/

#ifndef _ZIP_H
#define _ZIP_H

#include "Backup.h"
#include "syszlib.h"

typedef void (*CryptFunction)( UInt32 numBlocks, UInt8* dest, UInt8* src, UInt8* key, Boolean encrypt );
typedef void (*CryptCloseFunction)( void );

#define ENCRYPT true
#define DECRYPT false

#define MAX_KEY_LENGTH   16
#define MAX_BLOCK_LENGTH 16

typedef struct {
    CryptFunction Crypt;         
    CryptCloseFunction Close;
    UInt16        inBlock;
    UInt16        keySize;
    UInt16        blockSize;
    UInt8         key[ MAX_KEY_LENGTH ];
} CryptInfo;

typedef struct {
    z_stream zs;
    FileRef ref;
    UInt8*  buffer;
    UInt32  inBuffer;
    UInt32  bufferPos;
    UInt32  bufferSize;
    UInt32  totalUncompressedSize;
    UInt32  rawSize;
    UInt32  rawRead;
    UInt32  crc32;
    UInt32  trailerSize;
    UInt32  headerSize;
    UInt32  date;
    Int16   level;
    UInt8   compressed:1;
    UInt8   write:1;
    UInt8   flushed:1;
    UInt8   noMore:1;
    CryptInfo crypt;
    UInt8   cryptBlock[ MAX_BLOCK_LENGTH ];
    UInt16  totInCryptBlock;
    UInt16  posInCryptBlock;
} ZFile;

#define ZVFS_FLAG_CRC32    0x1

ZIP_SECTION void InitZip( void );
ZIP_SECTION void CleanupZip( void );
ZIP_SECTION Err ZVFSWriteClose( ZFile* zref );
ZIP_SECTION Err ZVFSReadClose( ZFile* zref );
ZIP_SECTION Err ZVFSWrite( ZFile* zref, UInt16 length, UInt8* s );
ZIP_SECTION Err ZVFSRead( ZFile* zref, UInt16 length, UInt8* s, UInt16* didReadP );
ZIP_SECTION Err ZVFSOpenRead( UInt16 vol, const Char* path, ZFile* zref, CryptInfo* cryptP );
ZIP_SECTION Err ZVFSOpenWrite( UInt16 vol, const Char* path, ZFile* zref, Int16 level, UInt32 mdate, CryptInfo* cryptP );
ZIP_SECTION Err ZVFSWrite( ZFile* zref, UInt16 length, UInt8* s );
ZIP_SECTION Err ZVFSRead( ZFile* zref, UInt16 length, UInt8* s, UInt16* didReadP );
ZIP_SECTION Err ZVFSCloseWrite( ZFile* zref );
ZIP_SECTION Err ZVFSCloseRead( ZFile* zref, Boolean check );
ZIP_SECTION Err ZVFSFileClose( ZFile* zref, Boolean check );
ZIP_SECTION Err ZVFSFileSize( ZFile* zref, UInt32* sizeP );
ZIP_SECTION Err ZVFSRewind( ZFile* zref );
ZIP_SECTION Err ZVFSStats( ZFile* zref, UInt32* date, UInt32* size, UInt32* crc32 );

#endif

